/* FILE: ListExpander.c
 *
 * PROGRAM: User-Friendly Metadata (UFM)
 *
 * DESCRIPTION: This file contains the functions that expand a component list,
 *     as read from an instruction file, into its internal, linked list
 *     representation.
 *
 * RCS INFO:
 *
 *      $Id: ListExpander.c,v 1.3 1999/04/06 15:36:40 ToolsDev Exp $
 *
 *      $Log: ListExpander.c,v $
 *      Revision 1.3  1999/04/06 15:36:40  ToolsDev
 *      Promotion of branch 1.2.0 (cleared wlint warnings) for release.
 *
 * Revision 1.2.0.1  1999/03/05  18:00:39  ToolsDev
 * Cleared several lint warnings.
 *
 * Revision 1.2  1999/02/03  16:23:18  stroup
 * Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.6  1999/01/26  18:21:01  stroup
 * Removed function ToUpStrcmp.
 *
 * Revision 1.1.0.5  1999/01/26  15:08:44  stroup
 * Moved SplitList to lstList.c and renamed to lstSplitList.  Changed
 * "SplitList" references to "lstSplitList".  Removed SplitList prototype.
 *
 * Revision 1.1.0.4  1999/01/22  15:43:35  stroup
 * Replaced calls to FreeAllList with calls to lstDestructor.  Removed function
 * FreeAllList and its prototype.  Made minor editing changes in functions
 * SplitList and AbsRefAdjust.
 *
 * Revision 1.1.0.3  1999/01/22  15:07:17  stroup
 * NewCompListElem calls odlMakeNode.  Added prototypes for local functions
 * SplitList, SyntaxChecker, CreateCompList, and NewCompListElem.
 *
 * Revision 1.1.0.2  1999/01/22  13:13:41  stroup
 * Moved RootNode, ParentNode, CompMatch, IsRootNode, OwnerAggregate, and
 * ValuesParameter to odlSupport.c and renamed.  Updated the calls to these
 * functions.  Deleted NextChildNode: call odlNextChildNode instead.  Removed
 * include of odldef.h.  Moved prototypes for IsRootNode, OwnerAggregate, and
 * ValuesParameter to odlSupport.h and renamed.
 *
 * Revision 1.1.0.1  1999/01/21  19:16:56  stroup
 * Moved NextToken, ltrim, rtrim, and strrev to strUtilities.c.  Moved their
 * corresponding prototypes to strUtilities.h.  Deleted obsolete and unused
 * version of JoinList.
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/

/* INCLUDED FILES:
*/
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "strUtilities.h"
#include "ListExpander.h"
#include "odlSupport.h"
#include "lstList.h"
#include "ufmMain.h"
#include "macMacro.h"

/* ---------------------------------------------------------------------*/

/* GLOBAL VARIABLES: the following variables may be used in other files.
 *    (global prototyes, if any, are given in the header file ListExpander.h)
*/
extern ufmControls ufmCONTROLS;

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
#define MSGLEN      256     /* maximum size of messages that can be
                               generated */

const char SLASH        = '/';     /* the forward slash character; used to
                                      separate components in a reference */
const char COMMA        = ',';     /* the comma character; used to separate
                                      references in a component list */
const char *CURR_COMP   = ".";     /* symbol in a reference that indicates
                                      "stay at the current node */
const char *PARENT_COMP = "..";    /* symbol in a reference that indicates "go
                                      to the parent node */
const char *ROOT_COMP   = "_ROOT"; /* symbol in a reference that indicates "go
                                      to the root node" */

char *curr_ref; /* Indicates the current reference being processed.  It is set
                   by ListExpander and used in messages to the user. */

/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static int SyntaxChecker (char *);
static lstLIST CreateCompList (ODLNodePtr, lstLIST);
static lstLIST NewCompListElem (lstLIST, ODLNodePtr);
static void AbsRefAdjust (lstLIST);
static void UserWarning (char *, char *);
static void UserError (char *, char *);
static void FatalError (char *, char *);

/* ---------------------------------------------------------------------*/

/*****************************************************************************
 * 
 * NAME: ListExpander
 *
 * DESCRIPTION: Part of what makes up a list instruction in an instruction
 *     file is the component list.  The component list identifies which
 *     components of an ODL label the list instruction applies to.  This list
 *     must be converted into an internal representation that is easy to use
 *     by the UFM tool--that is the job of the list expander function
 *     (ListExpander).  This function accepts two input arguments:
 *     1) a structure that points to a node in the ODL tree, i.e., the current
 *        node.  This is the node where traversal of the ODL tree begins for a
 *        relative reference.
 *     2) a component list.
 *
 *     Based on this input and the ODL tree itself, this function creates an
 *     internal representation (i.e., a data structure) of the given component
 *     list.  The data structure created, and returned, by this function is a
 *     linked list of pointers into the ODL tree.  There is one pointer for
 *     each component that the given component list resolves to.
 *
 *     In addition to its primary task of creating and returning the component
 *     list data structure, this function also performs several secondary
 *     tasks:
 *     1) macro substitution --  A reference in the given component list may
 *        contain macros.  In order to identify the component or components
 *        referred to by the reference, all macros must be resolved via string
 *        substitution.  Macro substitution must be done before the primary
 *        task and any of the other secondary tasks of this function can be
 *        performed.
 *     2) ODL tree navigation -- Using the macro-expanded references, this
 *        function navigates through the ODL tree taking into account sibling
 *        components of the same name.
 *     3) reference verification -- This function verifies that references are
 *        syntactically correct and that the referenced components actually
 *        exist in the ODL tree.
 *
 * PARAMETERS:
 *     curr_node (ODLNodePtr)  A structure that points to the current
 *                             component (or node).
 *     comp_list (char *)      A string of comma-delimited references.
 *
 * RETURN VALUE:
 *     ListExpander (lstLIST)  A linked list of structures, such that each
 *                             structure points to a node in the ODL tree.
 *                             This list is the internal representation of the
 *                             component list given as input.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Structure pointing to the current component and a component
 *              list string.
 *        POST: Function returns a linked list of structures, such that each
 *              structure points to a node in the ODL tree that one of the
 *              given references, in the component list, resolved to.  (See
 *              notes, below.)
 *     2. PRE:  Structure pointing to the current component and a component
 *              list string which is empty.
 *        POST: Function returns an empty list.
 *
 * NOTES:
 *     A reference in the given component list is ignored if any of the
 *     following conditions are true:
 *     1) the reference contains a syntax error (message generated,
 *        ReportWriter shut off);
 *     2) the reference resolves to a component in the ODL tree that does not
 *        exist (message generated only in verbose mode); or
 *     3) the reference, at some point during its resolution, refers to the
 *        parent of the root node (message generated, ReportWriter shut off).
 *
 * REVISION HISTORY:
 *     09/23/1998  J.D. Stroup  Original
 *     01/26/1999  J.D. Stroup  Changed "SplitList" references to
 *                              "lstSplitList".
 *
 ****************************************************************************/

lstLIST ListExpander (ODLNodePtr curr_node, char *comp_string)
{
  char    *comp_string_expand = comp_string;
  char    *reference;
  lstLIST comp_list, ref_list, qList, pList = lstInitList ();

  /* Expand all macros in the input component list (comp_list).  The input
     argument, comp_list, is a string containing zero or more comma-delimited
     references.  The function returns the component list with all macros
     expanded.  If the input string, comp_list, has no macros then the content
     of the returned string is exactly the same as the content of the input
     string. */

  comp_string_expand = macMacroExpander (comp_string);

  /* Split the macro-expanded component list into references.  The component
     list containing a string of comma-delimited, macro-expanded references is
     the input, along with the delimiter character itself.  The function
     returns a linked list.  Each element of the returned list contains one
     reference as given in the instruction file following the expansion of all
     user and predefined macros. */

  comp_list = lstSplitList (COMMA, comp_string_expand);

  /* For each reference that was split out of the macro-expanded component
     list... */

  while ((reference = (char *) lstIterator (comp_list)) != NULL)
  {
    curr_ref = reference;

    /* Check the reference's syntax.  If it contains an error, skip it. */

    if (! SyntaxChecker (reference)) continue;

    /* Convert the reference into a linked list.  Note that if an absolute
       reference is given, the first element of the resulting linked list
       contains an empty string and NOT the first component name.  It is the
       second element that contains the first component name. */

    ref_list = lstSplitList (SLASH, reference);

    /* Determine whether the reference is absolute or not.  If it is an
       absolute reference, then replace the empty string, in the first element
       of the linked list, (see above comment) with "_ROOT".  This helps later
       processing identify that this first element is referring to the root
       node of the ODL tree. */

    if (reference[0] == SLASH) AbsRefAdjust (ref_list);

    /* Beginning at the current node (curr_node), traverse the ODL tree
       identifying all components referred to by the reference.  A linked list
       of these components is created such that each element in the list is a
       pointer structure to the corresponding component node in the ODL
       tree. */

    qList = CreateCompList (curr_node, ref_list);

    /* During each iteration of this for-loop, the returned list from
       CreateCompList is appended to the component pointer structure list
       being created here in ListExpander. */

    lstAppendTo (pList, qList);
    qList = lstFreeList (qList);

    /* Free the reference list, including all the strings it points to. */

    ref_list = lstDestructor (ref_list, strDestructor);
  }

  /* Free the component list, including all the strings it points to. */

  comp_list = lstDestructor (comp_list, strDestructor);

  /* Return a pointer to the created component pointer structure list. */

  return (pList);

} /* end ListExpander */

/*****************************************************************************
 * 
 * NAME: SyntaxChecker
 *
 * DESCRIPTION: This function checks the syntax of the given reference.
 *
 * PARAMETERS:
 *     ref (char *)  Pointer to a string containing a reference.
 *
 * RETURN VALUE:
 *     SyntaxChecker (int)  0 --> Given reference contains a syntax error.
 *                          1 --> Given reference is valid: it does not
 *                                contain a syntax error.
 *
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Reference is syntactically valid.
 *        POST: Function returns 1.
 *     2. PRE:  Reference contains at least one syntax error.
 *        POST: Function generates a message, shuts off ReportWriter (via
 *              UserError), and returns 0.
 *
 * NOTES:
 *     This version of SyntaxChecker does not handle index specifications in a
 *     reference.  They will be trapped as syntax errors.  A future
 *     enhancement to this function will be the ability to recognize and
 *     handle index specifications.
 *
 * REVISION HISTORY:
 *     09/23/1998  J.D. Stroup  Original
 *
 ****************************************************************************/

static int SyntaxChecker (char *ref)
{
  char *token, *msg;
  char delimiter[2];
  int  ref_err = 0;
  unsigned int  token_len;
  int  comp_err, token_num;

  const char *Format1 = "reference contains a zero length component name.  "
                        "Reference is '%s', and is ignored";

  const char *Format2 = "reference contains an invalid component name, '%s'."
                        "  Reference is '%s'.  Component name must begin with"
                        " a letter, followed by 0 or more letters, digits, or"
                        " undersores, the last of which must be a letter or "
                        "digit.  The reference is ignored.";

  const char Underscore = '_';

  delimiter[0] = SLASH;
  delimiter[1] = '\0';

  /* Check each token, i.e., component name, of the reference. */

  for (token = strNextToken (ref, delimiter), token_num = 1;
       token != NULL;
       token = strNextToken (NULL, delimiter), token_num++)
  {
    comp_err = 0;

    /* If we've been passed an absolute reference (identified by its first
       character being a slash), then the first token (component name)
       returned by strNextToken is empty.  Therefore, we skip it. */

    if ((token_num == 1) && (ref[0] == SLASH))
      continue;

    /* Check its length: all component names must be greater than zero.  (See
       above comment for exception.) */

    token_len = strlen (token);
    if (token_len == 0)
    {
      unsigned int msglen = strlen (Format1) + strlen (curr_ref) + 1;

      comp_err = 1;

      msg = (char *) malloc (msglen * sizeof (char));

      if (msg == NULL)
        FatalError ("SyntaxChecker",
           "malloc failure: cannot allocate memory for error message string");

      sprintf (msg, Format1, curr_ref);
      UserError ("SyntaxChecker", msg);
      free (msg);
    }

    /* Check for the special "path" symbols, "." and "..".  Skip them. */

    else if ((strcmp (token, CURR_COMP)   == 0) ||
             (strcmp (token, PARENT_COMP) == 0))
      continue;

    /* Component name contains at least one character, and it is not one of
       the special "path" symbols. */

    else
    {
      /* Check the first character of the component name: it must be
         alphabetic. */

      if (! isalpha ( (int) token[0]))
        comp_err = 1;

      /* Check the last character of the component name: it must be alphabetic
         or numeric. */

      else if (! isalnum ( (int) token[token_len - 1]))
        comp_err = 1;

      /* Check the middle characters of the component name: they must be
         alphabetic, numeric, or the undersore. */

      else
      {
        int i;

        for (i = 1; i < (int) token_len - 1; i++)
          if ((! isalnum ( (int) token[i])) && (token[i] != Underscore))
            comp_err = 1;
      }

      /* If a syntax error was encountered, issue an error message and turn
         off the ReportWriter module (UserError). */

      if (comp_err)
      {
        unsigned int msglen = strlen (Format2) + strlen (curr_ref) +
                     strlen (token) + 1;

        msg = (char *) malloc (msglen * sizeof (char));

        if (msg == NULL)
          FatalError ("SyntaxChecker",
           "malloc failure: cannot allocate memory for error message string");

        sprintf (msg, Format2, token, curr_ref);
        UserError ("SyntaxChecker", msg);
        free (msg);
      }
    }
    free (token);
    if (ref_err == 0) ref_err = comp_err;
  }

  if (ref_err)
    return (0);
  else
    return (1);
} /* end SyntaxChecker */

/*****************************************************************************
 * 
 * NAME: AbsRefAdjust
 *
 * DESCRIPTION: For the given absolute reference, this function replaces its
 *              initial element containing--or whose data item is--the empty
 *              string with an initial element containing the string "_ROOT".
 *
 * PARAMETERS:
 *     ref_list (lstLIST)  Pointer to an absolute reference in internal,
 *                         linked list representation.
 *
 * RETURN VALUE:
 *     None.
 *
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Sufficient space is available on the heap.
 *        POST: Function replaces the first list element with one containing
 *              the string "_ROOT".
 *     2. PRE:  There is not sufficient space on the heap.
 *        POST: Function generates a fatal error message and terminates the
 *              program (via FatalError).
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/22/1998  J.D. Stroup  Original
 *     01/22/1999  J.D. Stroup  Removed the local void pointer vp.  Replaced
 *                              vp with root_sym in call to lstAddFirst.
 *
 ****************************************************************************/

static void AbsRefAdjust (lstLIST ref_list)
{
  char *root_sym = strdup (ROOT_COMP);

  if (root_sym == NULL)
    FatalError ("AbsRefAdjust",
                "malloc failure: cannot allocate memory for ROOT string");

  lstPluckFirst (ref_list);   /* remove first element */

  lstAddFirst (ref_list, (void *) root_sym); /* add new first element */
} /* end AbsRefAdjust */

/*****************************************************************************
 * 
 * NAME: CreateCompList
 *
 * DESCRIPTION: This function locates the component(s) in an ODL tree that the
 *     given reference resolves to.  The function requires two input
 *     arguments:
 *     1) a structure that points to a node in the ODL tree.  This is the
 *        beginning point for the ODL tree traversal.
 *     2) the linked list representation of a reference.  This argument can be
 *        thought of as the "roadmap" for traversing the ODL tree.  It
 *        indicates the path that the function needs to travel in order to get
 *        from the starting point to the final destination, which is a
 *        particular component--or even a set of components--in the ODL tree.
 *
 *     The function uses recursion to traverse the ODL tree.  It returns a
 *     linked list of pointers to the nodes (components) in the ODL tree
 *     referred to by the given reference.  If the given reference resolves to
 *     only one component, then a list containing one element is returned.  If
 *     the given reference resolves to no components, then a null list
 *     (pointer) is returned.  This latter case can occur when a reference
 *     resolves to an optional component that, in a particular instance, does
 *     not exist.  The returned list is referred to as the "component list".
 *
 * PARAMETERS:
 *     base_node (ODLNodePtr)  A structure pointing to a node in the ODL tree.
 *     reference (lstLIST)     Linked list representation of a reference.
 *                             Each node represents one component in the
 *                             reference.
 *
 * RETURN VALUE:
 *     CreateCompList (lstLIST)  A linked list of ODLNodePtr structures.  Each
 *                               structure is contained in one element of the
 *                               list and points to a node in the ODL tree.
 *
 * GLOBAL VARIABLES:
 *     curr_ref (char *)  A pointer to a string containing the reference
 *                        currently being processed.
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Base_node is pointing to a node in the ODL tree; reference
 *              resolves to a valid component, or components, in the ODL tree.
 *        POST: Function returns a component list where each element of the
 *              list is a structure pointing to the component, or components,
 *              referred to in the given reference.
 *     2. PRE:  Base_node is pointing to a node in the ODL tree; reference
 *              resolves to at least one component that is NOT found in the
 *              ODL tree.
 *        POST: Function returns a component list where each element of the
 *              list is a structure pointing to the component, or components,
 *              referred to in the given reference.  If all components
 *              referred to do not exist in the ODL tree, the returned list is
 *              empty.  A message is generated to a message file if the
 *              program is in verbose mode.
 *     3. PRE:  Base_node is pointing to a node in the ODL tree; reference, at
 *              some point, refers to the parent of the root node.
 *        POST: Function returns an empty component list.  A message is
 *              generated to a message file.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/21/1998  J.D. Stroup  Original
 *     01/21/1999  J.D. Stroup  Replaced calls to CompMatch, NextChildNode,
 *                              ParentNode, and RootNode with odlCompMatch,
 *                              odlNextChildNode, odlParentNode, and
 *                              odlRootNode due to these functions being
 *                              renamed.
 *     01/26/1999  J.D. Stroup  Changed "SplitList" references to
 *                              "lstSplitList"--this only occurred in
 *                              comments.
 *
 ****************************************************************************/

static lstLIST CreateCompList (ODLNodePtr base_node, lstLIST reference)
{
  lstLIST    plist, qlist;
  char       *comp_name;

  /* Initialize the component list, which is created and returned by this
     function.  Check the input reference: if it's null or empty, immediately
     return. */

  plist = lstInitList ();

  if ((reference == NULL) || (lstNumEl (reference) == 0)) return (plist);

  /* Get the next element (component) of the reference.  Once this element is
     retrieved, it becomes the current reference element. */

  comp_name = (char *)lstIterator (reference);

  /* If we've reached the end of the reference, then create a new element
     for--and add it to the end of--the component list.  Also, stop the
     iteration, so that the subsequent iterator call (which will be
     lstReverseIterator) works correctly (returning the last component of the
     reference). */

  if (comp_name == NULL)
  {
    plist = NewCompListElem (plist, base_node);
    lstStopIteration (reference);
  }

  /* If the current reference element refers to the current component, then
     re-invoke this function with the current node (base_node) and the
     reference. */

  else if (strcmp (comp_name, CURR_COMP) == 0)
  {
    qlist = CreateCompList (base_node, reference);
    lstAppendTo (plist, qlist);
    lstFreeList (qlist);
  }

  /* If the current reference element refers to the parent component... */

  else if (strcmp (comp_name, PARENT_COMP) == 0)
  {
	  /* Get the given node's (base_node's) parent. */

    ODLNodePtr parent_node = odlParentNode (base_node);

	  /* If the parent exists then re-invoke this function with the parent node
       and the reference. */

    if (parent_node.node_kind != UNDEFINED_NODE)
    {
      qlist = CreateCompList (parent_node, reference);
      lstAppendTo (plist, qlist);
      lstFreeList (qlist);
    }

	  else /* base_node is the root and we attempted to get its parent, which
            there is none */
    {
      char *format = "attempted to reference the parent of the root node "
                     "while processing reference '%s'.  The reference is "
                     "ignored.";
      char *msg = (char *) malloc ((strlen (format) + strlen (curr_ref) + 1) *
                                   sizeof (char));

      if (msg == NULL)
        FatalError ("CreateCompList",
           "malloc failure: cannot allocate memory for error message string");

      /* Issue a user error message and turn off the ReportWriter module.
         Note that an empty list will be returned by this call to
         CreateCompList. */

      sprintf (msg, format, curr_ref);

      UserError ("CreateCompList", msg);

      free (msg);
    }
  }

  /* If the current reference element refers to the root node... */

  else if (strcmp (comp_name, ROOT_COMP) == 0)
  {
    /* This situation refers solely to the case of an absolute reference.
       When a string containing an absolute reference is split into elements
       of a linked list (by lstSplitList, called from ListExpander), the first
       element of the list contains an empty string.  (This is by definition
       of lstSplitList, i.e., when the first character of the string given to
       lstSplitList is the delimiter character itself, then the first element
       created in the resulting linked list contains an empty string.)
       ListExpander replaces this empty string with the string "_ROOT".
       Therefore when CreateCompList encounters the string "_ROOT", it creates
       a structure pointing to the root node of the ODL tree (odlRootNode).
       CreateCompList is re-invoked and is passed this structure along with
       the referece. */

	  ODLNodePtr root_node = odlRootNode (base_node);
    qlist = CreateCompList (root_node, reference);
    lstAppendTo (plist, qlist);
    lstFreeList (qlist);
  }

  /* Otherwise, the current reference element refers to a component of the
     current node (base_node). */

  else
  {
    int        counter = 0;
    ODLNodePtr child_node;
    
    child_node.node_kind = UNDEFINED_NODE;
    child_node.node_type.undefined = NULL;

    child_node = odlNextChildNode (base_node, child_node);

    /* For each component of the current node (base_node) that matches the
       current reference element, re-invoke this function with the matching
       component and the reference.  The list returned from this call is
       appended to the component list being created by this function. */

    while (child_node.node_kind != UNDEFINED_NODE)
    {
      if (odlCompMatch (child_node, comp_name))
      {
        counter++;
        qlist = CreateCompList (child_node, reference);
        lstAppendTo (plist, qlist);

        /* Free qlist--but not the data it's pointing to--since the list is no
           longer needed: a copy was appended to plist. */

        qlist = lstFreeList (qlist);

        /* The prior call to CreateCompList, above, bumped the reference list
           iterator.  We need to move it back to its position before that
           call. */

        lstReverseIterator (reference);
      }
      child_node = odlNextChildNode (base_node, child_node);
    }

    /* If the current node had no matching components, then issue a message to
       the user.  (NOTE: the message is only generated if the verbose switch
       is set.) */

    if (counter == 0)
    {
      char *format = "component name, '%s', not found in the input ODL tree "
                     "while processing reference '%s'.  The reference is "
                     "ignored.";
      char *msg = (char *) malloc ((strlen (format) + strlen (comp_name) +
                                    strlen (curr_ref) + 1) * sizeof (char));

      if (msg == NULL)
        FatalError ("CreateCompList",
           "malloc failure: cannot allocate memory for error message string");

      sprintf (msg, format, comp_name, curr_ref);

      UserWarning ("CreateCompList", msg);

      free (msg);
    }
  }

  /* Return a pointer to the created component list. */

  return (plist);

} /* end CreateCompList */

/*****************************************************************************
 * 
 * NAME: NewCompListElem
 *
 * DESCRIPTION: This function creates a new element for the component pointer
 *     structure list (hereafter referred to as simply the component list).
 *     Its input is the component list and a structure pointing to a node in
 *     the ODL tree.  The information in the input structure is used to set
 *     the fields in the new element, after memory has been allocated for it.
 *     The function returns a pointer to the component list, now containing
 *     the new element.
 *
 * PARAMETERS:
 *     base_node (ODLNodePtr)  Structure pointing into the ODL tree used to
 *                             initialize the data contents of the component
 *                             list element created by this function.
 *     pList (lstLIST)         The list that the new element is going to be
 *                             added to.
 *
 * RETURN VALUE:
 *     NewCompListElem (lstLIST)  Pointer to a linked list that now includes a
 *                                new element added by this function.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *	
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/16/1998  J.D. Stroup  Original
 *     01/22/1999  J.D. Stroup  Replaced the code that allocated memory for
 *                              new element in component list, and initialized
 *                              it, with a call to odlMakeNode.
 *
 ****************************************************************************/

static lstLIST NewCompListElem (lstLIST pList, ODLNodePtr base_node)
{
  /* Create the new element for the component list.  Note that odlMakeNode is
     expecting a void pointer as its second argument.  This pointer is
     pointing at an aggregate node (AGGREGATE), parameter node (PARAMETER), or
     value node (VALUE).  This function must pass one of the node_type
     pointers of base_node.  The pointers of node_type are a union, all
     sharing the same memory location.  Since they're all pointers, they
     occupy the same amount of memory.  So instead of passing the pointer
     corresponding to base_node's node_kind, the "undefined" pointer is always
     passed.  The main reason for choosing the "undefined" pointer--as opposed
     to the "aggregate" pointer, for instance--is because it's a void pointer
     already. */

  ODLNodePtr *item = odlMakeNode (base_node.node_kind,
                                  base_node.node_type.undefined);

  /* Add the new element to the end of the given list. */

  lstAddLast (pList, (void *) item);

  /* Return the list with the new element added. */

  return (pList);

} /* end NewCompListElem */

/*****************************************************************************
 * 
 * NAME: UserWarning
 *
 * DESCRIPTION: Reports warnings if the ufmCONTROLS.verbose flag is set.
 *
 * PARAMETERS:  
 *     functionName (char *)   Pointer to a string containing the name of the
 *                             function that trapped the error.
 *     msg          (char *)   Pointer to a buffer containing a description of
 *                             the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pCompilerErrors (FILE)  The file to which messages to the
 *                                         user are to be written.
 *	
 * REVISION HISTORY:
 *   09/18/98  J.D. Stroup  Cloned from sltWarning
 *
 ****************************************************************************/

static void UserWarning (char *functionName, char *msg)
{
	if (!ufmCONTROLS.verbose) return;

  fprintf (ufmCONTROLS.pCompilerErrors, "%s: WARNING - %s\n",
                                        functionName, msg);

	return;

} /* end UserWarning */

/*****************************************************************************
 * 
 * NAME: UserError
 *
 * DESCRIPTION: Reports user errors.
 *
 * PARAMETERS:  
 *     functionName (char *)   Pointer to a string containing the name of the
 *                             function that trapped the error.
 *     msg          (char *)   Pointer to a buffer containing a description of
 *                             the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pCompilerErrors (FILE)  The file to which messages to the
 *                                         user are to be written.
 *	
 * REVISION HISTORY:
 *   09/19/98  J.D. Stroup  Cloned from UserWarning
 *
 ****************************************************************************/

static void UserError (char *functionName, char *msg)
{
  fprintf (ufmCONTROLS.pCompilerErrors, "%s: ERROR - %s\n",
                                        functionName, msg);

  ufmCONTROLS.CompilerFailure = 1; /* turn off ReportWriter module */

	return;

} /* end UserError */

/*****************************************************************************
 * 
 * NAME: FatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName (char *)  Pointer to a string containing the name of the
 *                            function that trapped the fatal error.
 *     msg          (char *)  Pointer to a buffer containing a description of
 *                            the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pDebugFile (FILE)  The file to which messages to the
 *                                    programmer are to be written.
 *	
 * REVISION HISTORY:
 *     09/16/98  J.D. Stroup  Cloned from lstFatalError.
 *
 ****************************************************************************/

static void FatalError (char *functionName, char *msg)
{
  fprintf (ufmCONTROLS.pDebugFile, "%s: FATAL - %s\n", functionName, msg);
	exit(-1);
} /* end FatalError */
