/* FILE: ODL_Node.h
 *
 * PROGRAM: User Friendly Metadata (UFM)
 *
 * DESCRIPTION: Whenever the UFM refers to a node in an ODL tree it does so
 *     via the data structure ODLNodePtr defined herein.  It provides a simple
 *     means of hiding the implementation of the ODL tree structure by
 *     providing a layer between the application's data structures and the ODL
 *     tree itself.  Through the use of this structure an application can
 *     refer (point) to any kind of node in an ODL tree.  The structure
 *     contains two fields.  The first field indicates the kind of node that
 *     the structure is pointing to.  It is an enumerated type that can take
 *     on one of four values: UNDEFINED, AGGREGATE, PARAMETER, and VALUE.  The
 *     second field is a union of four pointers.  The first field is used to
 *     interpret the union and know which pointer to follow.  The relationship
 *     of these two fields is illustrated below:
 *
 *                   ...then the pointer
 *     If node_kind  (in node_type) to
 *     is...         follow is...         ...and it is pointing to...
 *     ------------  -------------------  ----------------------------------
 *     UNDEFINED     undefined            null.
 *     AGGREGATE     aggregate            an aggregate node in the ODL tree.
 *     PARAMETER     parameter            a parameter node in the ODL tree.
 *     VALUE         value                a value node in the ODL tree.
 *
 *     An example of where this data structure is used is the internal
 *     representation of a component list.  The list itself is of type lstLIST
 *     (see lstList.h).  But the item, or data, that each element in the list
 *     is pointing to is one of these structures.  And, of course, each of
 *     these structures is in turn pointing to a particular node in the ODL
 *     tree.  Another example is the current component.  It can refer to any
 *     kind of node in the ODL tree, therefore, it is this structure as well.
 *
 * RCS INFO:
 *
 *	$Id: ODL_Node.h,v 1.2 1999/02/03 16:29:28 stroup Exp $
 *
 *	$Log: ODL_Node.h,v $
 *	Revision 1.2  1999/02/03 16:29:28  stroup
 *	Promotion of 1.1.0 branch.
 *
 * Revision 1.1.0.1  1999/01/20  19:30:07  stroup
 * Removed prototypes.
 *
 * Revision 1.1  1999/01/06  21:26:56  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/
#ifndef _ODL_Node_header
#define _ODL_Node_header

/* INCLUDED FILES: the following included files are here because of
 *     dependencies that exist within this header file.
*/
#include "odldef.h"
#include "lstList.h"


/* ---------------------------------------------------------------------*/

/* EXTERN VARIABLES: the following variables may be used in other files.
*/

enum Node_Kinds {UNDEFINED_NODE, AGGREGATE_NODE, PARAMETER_NODE, VALUE_NODE};

typedef enum Node_Kinds NODE_KIND;

struct ODL_Node
{
  NODE_KIND node_kind;
  union
  {
    void      *undefined;
    AGGREGATE aggregate;
    PARAMETER parameter;
    VALUE     value;
  } node_type;
};

typedef struct ODL_Node ODLNodePtr;

/* ---------------------------------------------------------------------*/

/* EXTERN PROTOTYPES:
*/
	/* NONE */

/* ---------------------------------------------------------------------*/

#endif 
