/* FILE: ReportWriter.c
 *
 * PROGRAM: User Friendly Metadata (UFM)
 *
 * DESCRIPTION: This file contains the functions that generate the ODL report
 *     in HTML format.
 *
 * RCS INFO:
 *
 * Revision 1.4  2002/07/11  13:40:00  ullman
 * Corrected tags to comply with HTML 4.01 Transitional
 *
 *      $Id: ReportWriter.c,v 1.3 2000/04/07 00:52:55 ToolsDev Exp $
 *
 *      $Log: ReportWriter.c,v $
 *      Revision 1.3  2000/04/07 00:52:55  ToolsDev
 *      Made changes to display the ODL file header informations:
 *      passed 3rd pointer for the header to HeaderOut function.
 *      added HTML tags to format the header informations.
 *      (Amer Rahabi)
 *
 *
 * Revision 1.2  1999/04/06  15:40:08  ToolsDev
 * Promotion of branch 1.1.0 (cleared wlint warnings) for release.
 *
 * Revision 1.1.0.2  1999/03/10  14:35:03  stroup
 * Further correction of errors found by lint: changed some variables of
 * type 'int' to 'unsigned int'.
 *
 * Revision 1.1.0.1  1999/03/06  19:00:26  ToolsDev
 * Mods for lint.
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/


/* INCLUDED FILES:
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


#include "ReportWriter.h"
#include "lstList.h"
#include "ufmMain.h"


/* ---------------------------------------------------------------------*/


/* GLOBAL VARIABLES: the following variables may be used in other files.
 *    (global prototyes, if any, are given in the header file ReportWriter.h)
*/
extern ufmControls ufmCONTROLS;


/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
const int SUCCESS = 1;
const int FAILURE = 0;
const int INDENTATION = 2; /* the number of spaces in the HTML source code
                              that each report level is indented from the
                              level above it */


typedef struct ColumnHeadingsAttributes
{
  char *cols_title;       /* columns title; spans all columns in table */
  char *cols_title_align; /* columns title alignment: "LEFT", "RIGHT",
                             "CENTER" */
  char **col_hdrs;        /* column headers; there are 'max_cols' column
                             headers */
  char *col_hdrs_align;   /* column header alignment: "LEFT", "RIGHT",
                             "CENTER" */
  int  rows_title_present; /* 1 --> there is a rows title for table; 0 --> no
                             row title */
  int  row_hdrs_present;  /* 1 --> there are row headers for table; 0 --> no
                             row headers */
  unsigned int  max_cols; /* total number of data columns in table */
} ColHdgsAttr;


typedef struct TableRowAttributes
{
  char *rows_title;       /* rows title; spans all data rows in table */
  char *rows_title_align; /* rows title alignment: "LEFT", "RIGHT", "CENTER" */
  char *row_hdr;         /* row headers; there are 'max_rows' row headers */
  char *row_hdr_align;   /* row header alignment: "LEFT", "RIGHT", "CENTER" */
  char **row_data;       /* data for the row; there are 'max_cols' pieces of
                            data */
  char *row_data_align;  /* row data alignment: "LEFT", "RIGHT", "CENTER" */
  int  first_row;        /* 1 --> this is the first row; 0 --> not the first
                            row */
  unsigned int max_rows; /* total number of data rows in table */
  unsigned int max_cols; /* total number of data columns in table */
} TblRowAttr;


typedef struct GeneralTableAttributes
{
  int border;       /* table border size: 0 --> no table border */
  int cell_padding; /* table cell padding: 0 --> no padding */
  int cell_spacing; /* table cell spacing: 0 --> no space between cells */
} GenTblAttr;


/* ---------------------------------------------------------------------*/


/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
/* ReportWriter functions */
static int HeaderOut (char *, char *, char *);  
static int TrailerOut (void);
static int SubListHandler (lstLIST, int);


/* SubListHandler functions */
static int StartList (int);
static int EndList (int);
static int SubListName (char *, int);
static int ParmHandler (char *, lstLIST, int);


static int TableHandler (struct reportTable *, int);


/* ParmHandler functions */
static int SingleValue (char *, char *, int);
static int MultValues (char *, lstLIST, int);
static int NoValues (char *, int);


/* TableHandler functions */
static int StartTable (GenTblAttr, int);
static int EndTable (int);
static int Caption (char *, int);
static int ColHeadings (ColHdgsAttr, int);
static char **GetRowData (unsigned int, unsigned int, char **);
static int TableDataRow (TblRowAttr, int);


/* Other general functions */
static char *AllocBlankSpace (int);
static void UserWarning (char *, char *);
static void UserError (char *, char *);
static void FatalError (char *, char *);


/* ---------------------------------------------------------------------*/


/*****************************************************************************
 * 
 * NAME: ReportWriter
 *
 * DESCRIPTION: This function generates an HTML report, representing an ODL
 *     label, based on a report structure that was created during the
 *     second-pass compiler step.  The creation of the report structure itself
 *     is based on a user-supplied ODL label and an optional instructions
 *     file.  The report structure is made up of report nodes.  Each node is a
 *     structure that points to one of three kinds of other data structures: a
 *     list (of type lstLIST), a parameter/values structure, and a table
 *     structure.  Generally a list represents the "children" (or
 *     substructures) of an ODL group or object, where each element of the
 *     list is a child of the group or object.  Each of the children may be
 *     another list, a parameter/values structure, or a table structure.  A
 *     parameter/values structure typically represents an ODL parameter and
 *     its value(s).  A table structure represents an ODL group or object that
 *     the user wishes to view in a table format.  This function returns an
 *     integer indication of whether it completed successfully or not.
 *
 * PARAMETERS:
 *     rep_struct (REPORT)  The report structure.  ReportWriter traverses this
 *                          structure and generates the HTML code.
 *
 * RETURN VALUE:
 *     ReportWriter (int)  SUCCESS --> This function completed successfully.
 *                         FAILURE --> This function encountered an
 *                                     error--most likely an I/O error--and
 *                                     could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


int ReportWriter (REPORT rep_struct)
{
  const int top_level = 0;


  /* Output HTML header information for the report.  This information is a
     title, a comment, and ODL file header information. these pieces of information are stored in
     fields of the given structure (rep_struct). */
  


  if (HeaderOut (rep_struct->pTitle, rep_struct->pComment,  rep_struct->pHeader) != SUCCESS)
    return (FAILURE);


    /* Output the ODL data in HTML format.  This "walks" the report structure,
       processing the sublists, parameter/values, and tables that make up this
       structure. */


  if (SubListHandler (rep_struct->pLevelOne, top_level) != SUCCESS)
    return (FAILURE);


  if (TrailerOut () != SUCCESS) return (FAILURE);


  return (SUCCESS);


} /* end ReportWriter */


/*****************************************************************************
 * 
 * NAME: HeaderOut
 *
 * DESCRIPTION: This function generates a title and comments in the output
 *     HTML report.
 *
 * PARAMETERS:
 *     title (char *)    The title of the HTML report.
 *     comment (char *)  Comments, i.e., a text string to be written at the
 *                       top of the HTML report.
 *
 *     header (char *)   Header in the odl file to be displayed after the optional user comment.
 *
 * RETURN VALUE:
 *     HeaderOut (int)  SUCCESS --> This function completed successfully.
 *                      FAILURE --> This function encountered an
 *                                  error--most likely an I/O error--and
 *                                  could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int HeaderOut (char *title, char *comment, char *header)
{
  fprintf (ufmCONTROLS.pHTML_Out, " <!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
  fprintf (ufmCONTROLS.pHTML_Out, "<HTML>\n");
  fprintf (ufmCONTROLS.pHTML_Out, "<HEAD>\n");
  if (title != NULL)
    fprintf (ufmCONTROLS.pHTML_Out, "<TITLE>%s</TITLE>\n", title);
  fprintf (ufmCONTROLS.pHTML_Out, "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">\n" );
  fprintf (ufmCONTROLS.pHTML_Out, "</HEAD>\n");
  fprintf (ufmCONTROLS.pHTML_Out, "<BODY>\n");
  if (comment != NULL)
    fprintf (ufmCONTROLS.pHTML_Out, "<I>%s</I>\n", comment);
fprintf (ufmCONTROLS.pHTML_Out, "<P>\n");
  if(header != NULL)
fprintf (ufmCONTROLS.pHTML_Out, "<PRE>%s</PRE>\n", header);
  fprintf (ufmCONTROLS.pHTML_Out, "<DL><DT><B>ECS Metadata ");
  if (title != NULL) fprintf (ufmCONTROLS.pHTML_Out, "for %s ", title);
  fprintf (ufmCONTROLS.pHTML_Out, "</B></DT>\n");



  return (SUCCESS);
} /* end HeaderOut */


/*****************************************************************************
 * 
 * NAME: TrailerOut
 *
 * DESCRIPTION: This function generates the "</body>" and "</html>" tags at
 *     the end of the HTML report.
 *
 * PARAMETERS:
 *     None.
 *
 * RETURN VALUE:
 *     TrailerOut (int)  SUCCESS --> This function completed successfully.
 *                       FAILURE --> This function encountered an
 *                                   error--most likely an I/O error--and
 *                                   could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int TrailerOut (void)
{
  fprintf (ufmCONTROLS.pHTML_Out, "</DL>\n</BODY>\n</HTML>\n");
  return (SUCCESS);
} /* end TrailerOut */


/*****************************************************************************
 * 
 * NAME: SubListHandler
 *
 * DESCRIPTION: This function handles and outputs--to the HTML report--a list
 *     of report nodes.  Each node in the list is a structure that points to
 *     another list (a sublist), a parameter/values structure, or a table
 *     structure.   If the report node is a sublist, then this function
 *     outputs the sublist's name at the current level, bumps the level, and
 *     recursively calls itself to handle the sublist.  (Note that with each
 *     invocation of this function, the level is increased.  When the function
 *     ends, the level returns to its original value before the call was
 *     made.)  If the report node is a parameter/values structure, then this
 *     function invokes ParmHandler to output the parameter and its value(s)
 *     at the current indentation level.  Finally, if the report node is a
 *     table structure, then this function outputs the table name at the
 *     current level, and invokes TableHandler to output an HTML table
 *     containing the data located in the table structure at the next level.
 *
 * PARAMETERS:
 *     sublist (lstLIST)  A list of report nodes.
 *     level (int)        The hierarchical level at which the given list is
 *                        output.
 *
 * RETURN VALUE:
 *     SubListHandler (int)  SUCCESS --> This function completed successfully.
 *                           FAILURE --> This function encountered an
 *                                       error--most likely an I/O error--and
 *                                       could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int SubListHandler (lstLIST rep_list, int level)
{
  REP_NODE rep_node;



  /* For each element (report node) in the given list... */


  lstStopIteration (rep_list);


  while ((rep_node = (REP_NODE) lstIterator (rep_list)) != NULL)
  {
    switch (rep_node->itemKind)
    {
      /* The report node is another sublist.  Output the sublist's name and
         the sublist itself at the next level. */


      case cmpSUBLIST:


        if (StartList (level + 1) != SUCCESS) return (FAILURE);
        if (SubListName (rep_node->pName, (level + 1)) != SUCCESS)
          return (FAILURE);
        if (SubListHandler ((lstLIST) rep_node->pItem, (level + 1)) != SUCCESS)
           return (FAILURE);
        if (EndList (level + 1) != SUCCESS) return (FAILURE);
        break;


      /* The report node is pointing to parameter/values data.  Output the
         data at the next level. */


      case cmpDATA:
        if (ParmHandler (rep_node->pName, (lstLIST) rep_node->pItem,
                         (level + 1)) != SUCCESS)
          return (FAILURE);
        break;


      /* The report node is pointing to table data.  Output the table's name
         only if one is given.  Output the table itself at the next level. */


      case cmpTABLE:
        if (strcmp (rep_node->pName, "") != 0)
          if (SubListName (rep_node->pName, (level + 1)) != SUCCESS)
            return (FAILURE);
        if (TableHandler ((struct reportTable *) rep_node->pItem,
                                                       (level + 1)) != SUCCESS)
          return (FAILURE);
        break;


      /* The report node is pointing to data that is to be ignored--skip
         it. */


      case cmpIGNORE:
        break;


      default:
        break;
    }
  }



  return (SUCCESS);
} /* end SubListHandler */


/*****************************************************************************
 * 
 * NAME: StartList
 *
 * DESCRIPTION: This function generates the necessary HTML tags to indicate
 *     that a new list is beginning.
 *
 * PARAMETERS:
 *     level (int)  The hierarchical level at which HTML by this function is
 *                  output.
 *
 * RETURN VALUE:
 *     StartList (int)  SUCCESS --> This function completed successfully.
 *                      FAILURE --> This function encountered an
 *                                  error--most likely an I/O error--and
 *                                  could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
*     None.
*
* NOTES:
*     None.
*
* REVISION HISTORY:
*     09/25/1998  J.D. Stroup  Original
*
****************************************************************************/


static int StartList (int level)
{
char *blanks = AllocBlankSpace (level * INDENTATION);


fprintf (ufmCONTROLS.pHTML_Out, "%s<DD><DL>\n", blanks);
free (blanks);


return (SUCCESS);
} /* end StartList */


/*****************************************************************************
* 
* NAME: EndList
*
* DESCRIPTION: This function generates the necessary HTML tags to indicate
*     the end of a list that began with the previous call to StartList.
*
* PARAMETERS:
*     level (int)  The hierarchical level at which HTML by this function is
*                  output.
*
* RETURN VALUE:
*     StartList (int)  SUCCESS --> This function completed successfully.
*                      FAILURE --> This function encountered an
*                                  error--most likely an I/O error--and
*                                  could not complete successfully.
*     
* GLOBAL VARIABLES:
*     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
*                                   written.
*       
* PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int EndList (int level)
{
  char *blanks = AllocBlankSpace (level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s</DL></DD>\n", blanks);
  free (blanks);


  return (SUCCESS);
} /* end EndList */


/*****************************************************************************
 * 
 * NAME: SubListName
 *
 * DESCRIPTION: This function generates the necessary HTML code indicating
 *     that the given name is that of a new list to be output.
 *
 * PARAMETERS:
 *     name (char *)  Name of the new list to be output.
 *     level (int)    The hierarchical level at which HTML by this function is
 *                    output.
 *
 * RETURN VALUE:
 *     SubListName (int)  SUCCESS --> This function completed successfully.
 *                        FAILURE --> This function encountered an
 *                                    error--most likely an I/O error--and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int SubListName (char *name, int level)
{
  char *blanks = AllocBlankSpace (level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s<DT><B>%s</B></DT>\n", blanks, name);
  free (blanks);


  return (SUCCESS);
} /* end SubListName */


/*****************************************************************************
 * 
 * NAME: ParmHandler
 *
 * DESCRIPTION: This function outputs a given parameter and its value(s) in
 *     HTML format to the report.  If the parameter has only one value, then
 *     it and its value are output in the following format:
 *
 *       parameter = value
 *
 *     If the parameter has more than one value, then its values are output as
 *     an ordered list according to the following format:
 *
 *       parameter =
 *         1. value 1
 *         2. value 2
 *         etc.
 *
 *     Although it's probably an error, this function does output a parameter
 *     that has no values.  In this case only the parameter name is written.
 *     An integer value is returned that indicates whether the function
 *     completed successfully or not.  Also, note that this function is passed
 *     the number of spaces to indent the HTML source code that it is
 *     generating.
 *
 * PARAMETERS:
 *     parm_name (char *)  The parameter's name.
 *     values (lstLIST)    A list containing the parameter's values.
 *     level (int)         The hierarchical level at which HTML by this
 *                         function is output.
 *
 * RETURN VALUE:
 *     ParmHandler (int)  SUCCESS --> This function completed successfully.
 *                        FAILURE --> This function encountered an
 *                                    error--most likely an I/O error--and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Parameter has only one value.
 *        POST: Function outputs a "parameter = value" format.
 *     2. PRE:  Parameter has more than one value.
 *        POST: Function outputs the parameter's values in an ordered list.
 *     3. PRE:  Parameter has no values.
 *        POST: Function outputs only the parameter's name.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/26/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int ParmHandler (char *parm_name, lstLIST values, int level)
{
  int num_vals = lstNumEl (values); /* Determine the number of values
                                       associated with the parameter. */
  char *value;


  /* The parameter has only one value.  Output the parameter and its value. */


  if (num_vals == 1)
  {
    lstStopIteration (values);
    value = (char *) lstIterator (values);
    if (SingleValue (parm_name, value, level) != SUCCESS)
      return (FAILURE);
  }


  /* The parameter has more than one value.  Output the parameter and its
     values as an ordered list. */


  else if (num_vals > 1)
  {
    if (MultValues (parm_name, values, level) != SUCCESS)
      return (FAILURE);
  }


  /* The parameter has no values (this is possibly an error).  Output the
     parameter name only. */


  else
    if (NoValues (parm_name, level) != SUCCESS) return (FAILURE);


  return (SUCCESS);
} /* end ParmHandler */


/*****************************************************************************
 * 
 * NAME: SingleValue
 *
 * DESCRIPTION: Given a parameter and its single value, this function
 *     generates the HTML for this list item.
 *
 * PARAMETERS:
 *     name (char *)   Name of the parameter.
 *     value (char *)  String representation of the parameter's value.
 *     level (int)     The hierarchical level at which HTML by this function
 *                     is output.
 *
 * RETURN VALUE:
 *     SingleValue (int)  SUCCESS --> This function completed successfully.
 *                        FAILURE --> This function encountered an
 *                                    error--most likely an I/O error--and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int SingleValue (char *name, char *value, int level)
{
  char *blanks = AllocBlankSpace (level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s<DD>%s = %s</DD>\n",
           blanks, name, value);
  free (blanks);


  return (SUCCESS);
} /* end SingleValue */


/*****************************************************************************
 * 
 * NAME: MultValues
 *
 * DESCRIPTION: Given a parameter with more than one value, this function
 *     generates the HTML for this list item.  Multiple values of a parameter
 *     are output via an ordered list.
 *
 * PARAMETERS:
 *     name (char *)     Name of the parameter.
 *     values (lstLIST)  A list of values.  Each element of the list points to
 *                       one value.
 *     level (int)       The hierarchical level at which HTML by this function
 *                       is output.
 *
 * RETURN VALUE:
 *     MultValues (int)  SUCCESS --> This function completed successfully.
 *                       FAILURE --> This function encountered an
 *                                   error--most likely an I/O error--and
 *                                   could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int MultValues (char *name, lstLIST values, int level)
{
  char *blanks1 = AllocBlankSpace (level * INDENTATION);
  char *blanks2 = AllocBlankSpace (INDENTATION);
  char *value;


  fprintf (ufmCONTROLS.pHTML_Out, "%s<DT>%s =\n", blanks1, name);
  fprintf (ufmCONTROLS.pHTML_Out, "%s<OL>\n", blanks1);


  lstStopIteration (values);


  while ((value = (char *) lstIterator (values)) != NULL)
    fprintf (ufmCONTROLS.pHTML_Out, "%s%s<LI>%s</LI>\n",
             blanks1, blanks2, value);


  fprintf (ufmCONTROLS.pHTML_Out, "%s</OL>\n%s</DT>\n", blanks1, blanks1);


  free (blanks1);
  free (blanks2);


  return (SUCCESS);
} /* end MultValues */


/*****************************************************************************
 * 
 * NAME: NoValues
 *
 * DESCRIPTION: Given a parameter that has no values, this function
 *     generates the HTML for this list item.
 *
 * PARAMETERS:
 *     name (char *)  Name of the parameter.
 *     level (int)    The hierarchical level at which HTML by this function is
 *                    output.
 *
 * RETURN VALUE:
 *     NoValues (int)  SUCCESS --> This function completed successfully.
 *                     FAILURE --> This function encountered an
 *                                 error--most likely an I/O error--and
 *                                 could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int NoValues (char *name, int level)
{
  char *blanks = AllocBlankSpace (level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s<DT>%s</DT>\n", blanks, name);
  free (blanks);


  return (SUCCESS);
} /* end NoValues */


/*****************************************************************************
 * 
 * NAME: TableHandler
 *
 * DESCRIPTION: This function outputs an HTML table containing the data stored
 *     in the given table structure to the report.  The following information
 *     is provided in the table structure:
 *
 *     1) caption for the table
 *     2) title of row data
 *     3) title of column data
 *     4) number of rows
 *     5) number of columns
 *     6) header for each row
 *     7) header for each column
 *     8) row and column data
 *
 *     All but the row and columns data are optional.
 *
 * PARAMETERS:
 *     tbl_data (struct reportTable *)
 *                        A structure containing all the data necessary to
 *                        generate an HTML table.
 *     level (int)        The hierarchical level at which the given table is
 *                        output.
 *
 * RETURN VALUE:
 *     TableHandler (int)  SUCCESS --> This function completed successfully.
 *                         FAILURE --> This function encountered an
 *                                     error--most likely an I/O error--and
 *                                     could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *     03/08/1999  J.D. Stroup  Removed unused local variable, "row_hdr", and
 *                              unused argument, "tbl_name".  Changed type of
 *                              variable, "i", from "int" to "unsigned int".
 *
 ****************************************************************************/


static int TableHandler (struct reportTable *tbl_data, int level)
{
  char *header_align  = "CENTER";
  char *data_align    = "CENTER";


  unsigned int  i;


  ColHdgsAttr col_hdg;
  TblRowAttr  row_info;
  GenTblAttr  table_info;


  /* Set up the local data structures:
       * general table attributes to pass to StartTable;
       * column headings attributes to pass to ColHeadings; and
       * table row attributes to pass to TableDataRow (Note that most of this
         structure's elements are fixed--i.e., set once upon entry to this
         function.  They are row independent.  But several are row dependent,
         and therefore change based on the row being output by TableDataRow
         (see below).) */


  table_info.border       = 1;
  table_info.cell_padding = 5;
  table_info.cell_spacing = 0;


  col_hdg.cols_title         = tbl_data->pColTitle;
  col_hdg.cols_title_align   = header_align;
  col_hdg.col_hdrs           = tbl_data->ppColHeaders;
  col_hdg.col_hdrs_align     = header_align;
  col_hdg.rows_title_present = (tbl_data->pRowTitle != NULL);
  col_hdg.row_hdrs_present   = (tbl_data->ppRowHeaders != NULL);
  col_hdg.max_cols           = tbl_data->numCols;


  row_info.rows_title       = tbl_data->pRowTitle;
  row_info.rows_title_align = header_align;
  row_info.row_hdr          = NULL; /* row dependent */
  row_info.row_hdr_align    = header_align;
  row_info.row_data         = NULL; /* row dependent */
  row_info.row_data_align   = data_align;
  row_info.first_row        = 1;
  row_info.max_cols         = tbl_data->numCols;
  row_info.max_rows         = tbl_data->numRows;


  /* Start the table, i.e., generate the <TABLE> tag. */


  if (StartTable (table_info, level) != SUCCESS)
    return (FAILURE);


  /* Generate the caption, if one is given. */


  if (Caption (tbl_data->pCaption, level) != SUCCESS)
    return (FAILURE);


  /* Generate the column headings. */


  if (ColHeadings (col_hdg, level) != SUCCESS) return (FAILURE);


  /* Generate the row data, including row title and row headers. */


  for (i = 0; i < tbl_data->numRows; i++)
  {
    /* Set the row dependent element of the table row attributes structure.
       These are the row's header string and the row's data. */


    row_info.row_hdr = (tbl_data->ppRowHeaders == NULL) ?
                           NULL                         :
                           tbl_data->ppRowHeaders[i];


    row_info.row_data = GetRowData (i, tbl_data->numCols, tbl_data->ppTable);


    /* Output the row. */


    if (TableDataRow (row_info, level) != SUCCESS) return (FAILURE);


    /* Perform clean up activities. */


    free (row_info.row_data);
    row_info.first_row = 0;
  }


  /* End the table, i.e., generate the </TABLE> tag. */


  if (EndTable (level) != SUCCESS) return (FAILURE);


  return (SUCCESS);
} /* end TableHandler */


/*****************************************************************************
 * 
 * NAME: StartTable
 *
 * DESCRIPTION: This function generates the <TABLE> tag to the HTML output
 *     report.  This tag includes attributes described in the given structure.
 *
 * PARAMETERS:
 *     table_info (GenTblAttr)  A structure containing general attributes for
 *                              the <TABLE> tag.
 *     tbl_level (int)          The hierarchical level at which the given
 *                              table is output.
 *
 * RETURN VALUE:
 *     StartTable (int)  SUCCESS --> This function completed successfully.
 *                       FAILURE --> This function encountered an
 *                                   error--most likely an I/O error--and
 *                                   could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int StartTable (GenTblAttr table_info, int tbl_level)
{
  char *blanks = AllocBlankSpace (tbl_level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out,
           "%s<DD><TABLE BORDER=%d CELLPADDING=%d CELLSPACING=%d>\n",
           blanks, table_info.border, table_info.cell_padding,
           table_info.cell_spacing);


  free (blanks);


  return (SUCCESS);


} /* end StartTable */


/*****************************************************************************
 * 
 * NAME: EndTable
 *
 * DESCRIPTION: This function generates the </TABLE> tag to the HTML output
 *     report.
 *
 * PARAMETERS:
 *     tbl_level (int)          The hierarchical level at which the given
 *                              table is output.
 *
 * RETURN VALUE:
 *     EndTable (int)  SUCCESS --> This function completed successfully.
 *                     FAILURE --> This function encountered an
 *                                 error--most likely an I/O error--and
 *                                 could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int EndTable (int tbl_level)
{
  char *blanks = AllocBlankSpace (tbl_level * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s</TABLE><DD>\n", blanks);
  free (blanks);


  return (SUCCESS);


} /* end EndTable */


/*****************************************************************************
 * 
 * NAME: Caption
 *
 * DESCRIPTION: This function generates a table's caption, via the
 *     <CAPTION>/</CAPTION> tags, to the HTML output report.
 *
 * PARAMETERS:
 *     caption (char *)  The table's caption.
 *     align (char *)    Value for the caption's align attribute.
 *     tbl_level (int)   The hierarchical level at which the given table is
 *                       output.
 *
 * RETURN VALUE:
 *     Caption (int)  SUCCESS --> This function completed successfully.
 *                    FAILURE --> This function encountered an
 *                                error--most likely an I/O error--and
 *                                could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static int Caption (char *caption, int tbl_level)
{
  if (caption != NULL)
  {
    char *blanks = AllocBlankSpace ((tbl_level + 1) * INDENTATION);


    fprintf (ufmCONTROLS.pHTML_Out, "%s<CAPTION>%s</CAPTION>\n",
             blanks, caption);


    free (blanks);
  }


  return (SUCCESS);


} /* end Caption */


/*****************************************************************************
 * 
 * NAME: ColHeadings
 *
 * DESCRIPTION: This function generates the HTML code for the column header
 *     rows of a table.  A table can have at most two column header rows: one
 *     containing the columns title--which spans all columns in the table, and
 *     one containing the column headers--one for each column in the table.
 *     Both of these pieces of column header information are optional.  Thus
 *     if no column header information is given, then this function outputs
 *     nothing.
 *
 * PARAMETERS:
 *     col_hdg (ColHdgsAttr)  A structure containing all the necessary
 *                            elements for the generation of HTML code for the
 *                            table heading.
 *     tbl_level (int)        The hierarchical level at which the given table
 *                            is output.
 *
 * RETURN VALUE:
 *     ColHeadings (int)  SUCCESS --> This function completed successfully.
 *                        FAILURE --> This function encountered an
 *                                    error--most likely an I/O error--and
 *                                    could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  Both the columns title and column headers are given.
 *        POST: The function outputs two table rows: one containing the
 *              columns title, and one containing the column headers.
 *     2. PRE:  Only the columns title is given.
 *        POST: The function outputs a table row containing the columns title.
 *     3. PRE:  Only the column headers are given.
 *        POST: The function outputs a table row containing the column
 *              headers.
 *     4. PRE:  Neither the columns title nor the column headers are given.
 *        POST: The function outputs nothing.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/28/1998  J.D. Stroup  Original
 *     03/09/1999  J.D. Stroup  Changed type of local variable "i" from "int"
 *                              to "unsigned int".  Changed fprintf statements:
 *                              "%d" to "%u".
 *
 ****************************************************************************/


static int ColHeadings (ColHdgsAttr col_hdg, int tbl_level)
{
  int colspan = 0;
  int rowspan = 0;


  char *tr_blanks = AllocBlankSpace ((tbl_level + 1) * INDENTATION);
  char *td_blanks = AllocBlankSpace ((tbl_level + 2) * INDENTATION);


  /* Based on the amount of row information and column information that has
     been given, determine the number of rows and columns that the empty cell
     in upper, lefthand corner of the table spans.  If no row information is
     given, or no column information is given, then there is no such empty
     cell in this table. */


  if (col_hdg.rows_title_present) colspan++;
  if (col_hdg.row_hdrs_present)   colspan++;
  if (col_hdg.cols_title != NULL) rowspan++;
  if (col_hdg.col_hdrs  != NULL)  rowspan++;


  /* No column information given. */


  if (rowspan == 0)
  {
    /* DONE..there's nothing to output. */
  }


  /* Some column information given. */


  else if (rowspan == 1)
  {
    /* Generate a row for either the columns title or column headers,
       depending on which one exists. */


    fprintf (ufmCONTROLS.pHTML_Out, "%s<TR>\n", tr_blanks); /* start row */


    /* Generate a blank cell in the upper lefthand corner of the table, if
       necessary--i.e., if some or all row information given. */


    if (colspan > 0)
      fprintf (ufmCONTROLS.pHTML_Out, "%s<TD COLSPAN=%d ROWSPAN=%d></TD>\n",
               td_blanks, colspan, rowspan);


    if (col_hdg.cols_title != NULL) /* columns title exists */
      fprintf (ufmCONTROLS.pHTML_Out, "%s<TH COLSPAN=%u ALIGN=%s>%s</TH>\n",
               td_blanks, col_hdg.max_cols, col_hdg.cols_title_align,
               col_hdg.cols_title);


    else /* column headers exist */
    {
      unsigned int i;
      for (i = 0; i < col_hdg.max_cols; i++)
        fprintf (ufmCONTROLS.pHTML_Out, "%s<TH ALIGN=%s>%s</TH>\n",
                 td_blanks, col_hdg.col_hdrs_align, col_hdg.col_hdrs[i]);
    }


    fprintf (ufmCONTROLS.pHTML_Out, "%s</TR>\n", tr_blanks); /* end row */
  }


  /* All column information given. */


  else /* rowspan = 2 */
  {
    unsigned int i;


    /* Generate a row for the columns title. */


    fprintf (ufmCONTROLS.pHTML_Out, "%s<TR>\n", tr_blanks); /* start row */


    /* Generate a blank cell in the upper lefthand corner of the table, if
       necessary--i.e., if some or all row information given. */


    if (colspan > 0)
      fprintf (ufmCONTROLS.pHTML_Out, "%s<TD COLSPAN=%d ROWSPAN=%d></TD>\n",
               td_blanks, colspan, rowspan);


    fprintf (ufmCONTROLS.pHTML_Out, "%s<TH COLSPAN=%u ALIGN=%s>%s</TH>\n",
             td_blanks, col_hdg.max_cols, col_hdg.cols_title_align,
             col_hdg.cols_title);


    fprintf (ufmCONTROLS.pHTML_Out, "%s</TR>\n", tr_blanks); /* end row */


    /* Generate a row for the column headers. */


    fprintf (ufmCONTROLS.pHTML_Out, "%s<TR>\n", tr_blanks); /* start row */


    for (i = 0; i < col_hdg.max_cols; i++)
      fprintf (ufmCONTROLS.pHTML_Out, "%s<TH ALIGN=%s>%s</TH>\n",
               td_blanks, col_hdg.col_hdrs_align, col_hdg.col_hdrs[i]);


    fprintf (ufmCONTROLS.pHTML_Out, "%s</TR>\n", tr_blanks); /* end row */
  }


  free (tr_blanks);
  free (td_blanks);


  return (SUCCESS);


} /* end ColHeadings */


/*****************************************************************************
 * 
 * NAME: GetRowData
 *
 * DESCRIPTION: This function retrieves a row of table data to be displayed.
 *     The row is identified by the given row number argument.
 *
 * PARAMETERS:
 *     row_num (int)         The number of the row in "table_data" to
 *                           retrieve.
 *     max_col (int)         The total number of columns of data in
 *                           "table_data".
 *     table_data (char **)  An array of pointers to the table data.
 *
 * RETURN VALUE:
 *     GetRowData (char **)  An array of pointers to strings.  The array
 *                           points to a copy of the requested row from the
 *                           input array, "table_data".
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  There is sufficient memory to allocate the strings.
 *        POST: A pointer to the allocated memory is returned.
 *     2. PRE:  There is insufficient memory to allocate the strings.
 *        POST: An error message is generated and the program terminates with
 *              a non-zero error status via FatalError.
 *
 * NOTES:
 *     The returned array of strings are allocated off of the heap via strdup.
 *
 * REVISION HISTORY:
 *     09/28/1998  J.D. Stroup  Original
 *     10/20/1998  J.D. Stroup  Removed references to row or column major.
 *     03/08/1999  J.D. Stroup  Removed unused argument, "max_row".
 *
 ****************************************************************************/


static char **GetRowData (unsigned int row_num, unsigned int max_col,
                          char **table_data)
{
  char **row_data = calloc (max_col, sizeof (char *));
  unsigned int  i, pos;


  for (i = 0; i < max_col; i++)
  {
    pos = (row_num * max_col) + i;


    row_data[i] = strdup (table_data[pos]);


    if (row_data[i] == NULL)
      FatalError ("GetRowData",
               "malloc failure in strdup: cannot allocate memory for string");
  }


  return (row_data);


} /* end GetRowData */


/*****************************************************************************
 * 
 * NAME: TableDataRow
 *
 * DESCRIPTION: This function generates the HTML code for one data row of the
 *     table.  The HTML code is written to the HTML output report.
 *
 * PARAMETERS:
 *     row_info (TblRowAttr)  A structure containing various information
 *                            pertaining to the content and format of the
 *                            table row to be displayed.
 *     tbl_level (int)        The hierarchical level at which the given
 *                            table is output.
 *
 * RETURN VALUE:
 *     TableDataRow (int)  SUCCESS --> This function completed successfully.
 *                         FAILURE --> This function encountered an
 *                                     error--most likely an I/O error--and
 *                                     could not complete successfully.
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pHTML_Out (FILE)  The file to which the HTML report is
 *                                   written.
 *      
 * PRE- & POST-CONDITIONS:
 *     None.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *     03/09/1999  J.D. Stroup  Changed type of local variable "i" from "int"
 *                              to "unsigned int".  Changed fprintf statements:
 *                              "%d" to "%u".
 *
 ****************************************************************************/


static int TableDataRow (TblRowAttr row_info, int tbl_level)
{
  unsigned int i;


  char *tr_blanks = AllocBlankSpace ((tbl_level + 1) * INDENTATION);
  char *td_blanks = AllocBlankSpace ((tbl_level + 2) * INDENTATION);


  fprintf (ufmCONTROLS.pHTML_Out, "%s<TR>\n", tr_blanks);


  /* If this is the first call to TableDataRow for this table--i.e., we're
     about to output the first row--and a row title is given, then generate
     the rows title cell, which spans all table data rows. */


  if (row_info.first_row && (row_info.rows_title != NULL))
    fprintf (ufmCONTROLS.pHTML_Out, "%s<TH ALIGN=%s ROWSPAN=%u>%s</TH>\n",
             td_blanks, row_info.rows_title_align, row_info.max_rows,
             row_info.rows_title);


  /* If a row header given, then generate its cell. */


  if (row_info.row_hdr != NULL)
    fprintf (ufmCONTROLS.pHTML_Out, "%s<TH ALIGN=%s>%s</TH>\n",
             td_blanks, row_info.row_hdr_align, row_info.row_hdr);


  /* Generate the row of table data. */


  for (i = 0; i < row_info.max_cols; i++)
    fprintf (ufmCONTROLS.pHTML_Out, "%s<TD ALIGN=%s>%s</TD>\n",
             td_blanks, row_info.row_data_align, row_info.row_data[i]);


  fprintf (ufmCONTROLS.pHTML_Out, "%s</TR>\n", tr_blanks); /* end row */


  return (SUCCESS);
} /* end TableDataRow */


/*****************************************************************************
 * 
 * NAME: AllocBlankSpace
 *
 * DESCRIPTION: This function allocates memory for a string that contains all
 *     spaces (blank characters).  The size of the string is determined by the
 *     value passed in as input to this function.
 *
 * PARAMETERS:
 *     num_blanks (int)  The number of blanks characters to allocate space
 *                       for.
 *
 * RETURN VALUE:
 *     AllocBlankSpace (char *)  A pointer to a memory location containing a
 *                               string of blank characters.  (See the PRE- &
 *                               POST-CONDITIONS, below.)
 *     
 * GLOBAL VARIABLES:
 *     None.
 *      
 * PRE- & POST-CONDITIONS:
 *     1. PRE:  There is sufficient memory to allocate the string.
 *        POST: A pointer to the allocated memory is returned.
 *     2. PRE:  There is insufficient memory to allocate the string.
 *        POST: An error message is generated and the program terminates with
 *              a non-zero error status via FatalError.
 *
 * NOTES:
 *     None.
 *
 * REVISION HISTORY:
 *     09/25/1998  J.D. Stroup  Original
 *
 ****************************************************************************/


static char *AllocBlankSpace (int num_blanks)
{
  char *blanks = (char *) malloc ((unsigned int) (num_blanks + 1) *
                                  sizeof (char));
  int  i;


  /* If the memory allocation failed, generate an error message and terminate
     the program. */


  if (blanks == NULL)
    FatalError ("AllocBlankSpace",
                "malloc failure: cannot allocate memory for string");


  /* Fill the allocated space with spaces. */


  for (i = 0; i < num_blanks; i++)
    blanks[i] = ' ';
  blanks[num_blanks] = '\0';


  return (blanks);
} /* end AllocBlankSpace */


/*****************************************************************************
 * 
 * NAME: UserWarning
 *
 * DESCRIPTION: Reports warnings if the ufmCONTROLS.verbose flag is set.
 *
 * PARAMETERS:  
 *     functionName (char *)   Pointer to a string containing the name of the
 *                             function that trapped the error.
 *     msg          (char *)   Pointer to a buffer containing a description of
 *                             the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pCompilerErrors (FILE)  The file to which messages to the
 *                                         user are to be written.
 *      
 * REVISION HISTORY:
 *   09/18/98  J.D. Stroup  Cloned from sltWarning
 *
 ****************************************************************************/


static void UserWarning (char *functionName, char *msg)
{
        if (!ufmCONTROLS.verbose) return;


  fprintf (ufmCONTROLS.pCompilerErrors, "%s: WARNING - %s\n",
                                        functionName, msg);


  /* Flush buffer contents so that memory allocated for the message can be 
     freed following the call to this function. */


  fflush (ufmCONTROLS.pCompilerErrors);


        return;


} /* end UserWarning */


/*****************************************************************************
 * 
 * NAME: UserError
 *
 * DESCRIPTION: Reports user errors.
 *
 * PARAMETERS:  
 *     functionName (char *)   Pointer to a string containing the name of the
 *                             function that trapped the error.
 *     msg          (char *)   Pointer to a buffer containing a description of
 *                             the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pCompilerErrors (FILE)  The file to which messages to the
 *                                         user are to be written.
 *      
 * REVISION HISTORY:
 *   09/19/98  J.D. Stroup  Cloned from UserWarning
 *
 ****************************************************************************/


static void UserError (char *functionName, char *msg)
{
  fprintf (ufmCONTROLS.pCompilerErrors, "%s: ERROR - %s\n",
                                        functionName, msg);


  /* Flush buffer contents so that memory allocated for the message can be 
     freed following the call to this function. */


  fflush (ufmCONTROLS.pCompilerErrors);


  ufmCONTROLS.CompilerFailure = 1; /* turn off ReportWriter module */


        return;


} /* end UserError */


/*****************************************************************************
 * 
 * NAME: FatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName (char *)  Pointer to a string containing the name of the
 *                            function that trapped the fatal error.
 *     msg          (char *)  Pointer to a buffer containing a description of
 *                            the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     ufmCONTROLS.pDebugFile (FILE)  The file to which messages to the
 *                                    programmer are to be written.
 *      
 * REVISION HISTORY:
 *     09/16/98  J.D. Stroup  Cloned from lstFatalError.
 *
 ****************************************************************************/


static void FatalError (char *functionName, char *msg)
{
  fprintf (ufmCONTROLS.pDebugFile, "%s: FATAL - %s\n", functionName, msg);
        exit(-1);
} /* end FatalError */

