/* FILE: macMarcro.c
 *
 * PROGRAM: User Friendly Metadata
 *
 * DESCRIPTION: This file contains the functions that provide the UFM tool
 *		with the ability to handle user-defined and predefined macros.
 *
 * RCS INFO:
 *
 *      $Id: macMacro.c,v 1.1 2002/05/08 15:20:20 yin Exp $
 *
 *      $Log: macMacro.c,v $
 *      Revision 1.1  2002/05/08 15:20:20  yin
 *      Initial revision
 *
 *      Revision 1.3  1999/04/06 15:50:22  ToolsDev
 *      Promotion of branch 1.2.0 (cleared wlint warnings) for release.
 *
 * Revision 1.2.0.2  1999/03/05  20:51:55  ToolsDev
 * More corrections for lint.
 *
 * Revision 1.2.0.1  1999/02/17  15:48:28  ToolsDev
 * Minor corrections to eliminate Lint warnings.
 *
 * Revision 1.2  1999/02/03  18:34:26  jarva
 * Promotions of the 1.1.0 branch
 *
 * Revision 1.1.0.4  1999/02/02  21:23:43  jarva
 * Upgrade for predefined macros with arguments.
 *
 * Revision 1.1.0.3  1999/01/26  17:16:12  jarva
 * Eliminated references to cmpComponentName.
 *
 * Revision 1.1.0.2  1999/01/21  22:31:48  jarva
 * *** empty log message ***
 *
 * Revision 1.1  1999/01/06  20:34:20  ToolsDev
 * Initial revision from prototype baseline.
 *
 *
*/

/* INCLUDED FILES:
*/
#include "macMacro.h"
#include "lstList.h"
#include "sltLUT.h"
#include "ufmMain.h"
#include "odlSupport.h"
#include "ListExpander.h"

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>

/* ---------------------------------------------------------------------*/

/* GLOBAL VARIABLES: the following variables may be used in other files.
 *    (global prototyes, if any, are given in the header file macMarcro.h)
*/
	/* NONE */
/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
static lstLIST pExpStack = lstLISTNULL; /* The expansion stack.  This is a
	* list of all the nested macros encountered during an expansion.
	* It is used to prevent infinite looping that would otherwise occur
	* because of self-referential macros.  The underlying list is 
	* initialized by macPushExpStack, if it hasn't been already.  That
	* function and its brother are the only ones that add or remove
	* elements from the stack.  Note that the underlying list contains
	* only strings; hence the special purpose function lstIsStrInList()
	* applies and is used to determine whether a given macro already
	* appears in the expansion stack. */
#define VERBOSE_DEBUG 1
#define VERBOSE 1
/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static int macExtractMacro( char *pString, char *pMacroName, 
	unsigned int *pStartRef, unsigned int *pEndRef, char **ppArg ); 
static void macUserError( char *functionName, char *macroName, char *msg );
static void macUserWarning( char *functionName, char *msg );
static void macFatalError( char *functionName, char *msg );
static void macDebugMsg( char *functionName, char *msg );
char *macGetMacroString( char *pMacro, char *pArg );
static char *macExpandString( char *pOldString, char *pMacroString,
	unsigned int startRef, unsigned int endRef );
char *macChildList( struct ODL_Node *pCompo );
/* ---------------------------------------------------------------------*/

/* EXTERNAL VARIABLES:
*/
extern sltLUT pMacroLUT; /* the LUT for macro expansions */
extern sltLUT pIDTranslationLUT; /* the LUT for translation
								  * strings for the ODL identifiers */
extern ufmControls ufmCONTROLS;
extern lstLIST pCCStack; /* the current component stack */

/*****************************************************************************
 * 
 * NAME: macMacroExpander
 *
 * DESCRIPTION: Expands the given string; that is, it replaces all macro
 *		references in the string with their replacement strings, and it
 *      replaces any macro references in the replacement strings with
 *      their replacement strings and so on until no macro references
 *      remain.
 *
 *		Macros that refer to themselves directly or indirectly are 
 *      treated as user errors. An error message is written and the
 *      CompilerFailure indicator gets set.  A pointer to a string on
 *      the heap will still be returned, but it won't be what was
 *      expected.
 *
 * PARAMETERS:  
 *     pString -- (char *) Points to a string containing 0 or more
 *				           macro references.
 *
 * RETURN VALUE:
 *     Pointer to a new string that has been allocated on the heap. This is
 *     true under all circumstances, even if no macro substitution is 
 *     performed.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: a string with no embedded macros is given. POST: a pointer to
 *        a duplicate of the original string is returned.
 *     2. PRE: a string with 1 or more user-defined macros is given. POST:
 *        Macros that can be found in the lookup table have been replaced
 *        by the associated translation string; macros that could not be
 *        found have been removed from the return string.
 *     3. PRE: a string with 1 or more predefined macros is given. POST: the
 *        predefined macros have been replaced with the appropriate string.
 *
 * NOTES:
 *     1. This function employs recursion.
 *     2. This fucntion employs a stack to keep track of nested macros in
 *        order to detect self-refential macors. See pExpStack.
 *     3. Prior to mods made on 01/26/99, it was considered an error for
 *        the predefined macro _VALUE to be processed by this code. That no
 *        longer applies.
 *
 * REVISION HISTORY:
 *   09/16/98  Ken Jarva  -- Origninal
 *   01/26/99  Ken Jarva  -- code for predefined macros with arguments.
 *
 ****************************************************************************/
char *macMacroExpander( char *pString )
{
	char *pExpString;
	char MacroName[ MAX_MACRO_NAME +1 ];
	char *pEmpty;
	char *pSubStr;
	char *pArg; /* points at macro's agrument buffer (if any) allocated by
				 * macExtractMacro. */
	unsigned int startRef, endRef;
	int status;

	/* The first time through, the underlying list of the expansion stack
	 * must be initialized. */
	if( pExpStack == lstLISTNULL ) pExpStack = lstInitList();

	/* Duplicate the input string: */
	pExpString = strdup( pString );

	/* Don't bother with empty strings...*/
	if( strlen(pExpString) == 0 ) return pExpString;

	/* Get the first macroc reference, if any, from the expansion 
	 *     string: */
	status = macExtractMacro(pExpString, MacroName, &startRef, &endRef, &pArg);

	/* Loop while there are macro references to expand...*/
	while( status == 1 ) {
		/* There is a macro to expand...*/
		/*     Is it already on the expansion stack? */
		if( lstIsStrInList(pExpStack, MacroName) ) {
			/* There is a self-referential macro. */
			macUserError( "macExpandMacro", MacroName, 
				"self-referential macro." );
			pEmpty = strdup( "" );
			free( pExpString );
			return pEmpty;
		}

		/* Push the macro name onto the expansion stack...*/
		lstAddFirst( pExpStack, MacroName );
		pSubStr = macGetMacroString( MacroName, pArg ); /* returns ptr into the 
		                                           * heap */
		if( pArg ) free( pArg );

		/* Expand the string; i.e., replace the macro reference with the
		 * appropriate string. */
		if( pSubStr ) {
			char *pStr; /* Temporary holder */
			pStr = macMacroExpander( pSubStr ); /* clear any nested macro
												 * references */
			free( pSubStr );
			pSubStr = pStr; /* this is the clean substitution string */

			/* Now substitute the replacement string for the macro ref. */
			pStr = macExpandString( pExpString, pSubStr, startRef, endRef );
			free( pExpString );
			free( pSubStr );
			pExpString = pStr;
		}

		/* Pop the expansion stack...*/
		(void) lstPluckFirst( pExpStack );

		status = macExtractMacro(pExpString, MacroName, &startRef, &endRef, &pArg);
	} /* end while */

	if( status == -1 ){
		/* There was an error in macExtractMacro. */
		free( pExpString );
		return strdup("");
	}

	return pExpString;
} /* end macMacroExpander */

/*****************************************************************************
 * 
 * NAME: macExtractMacro
 *
 * DESCRIPTION: Searches the given string for the first macro reference, if
 *		any.  The name of the found macro and the relative indexes of the
 *      first and last characters of the reference are returned through the
 *      calling arguments.
 *		
 *
 * PARAMETERS:  
 *     pString      -- (char *) Points to the string that's to be expanded.
 *     pMacroName   -- (char *) Points to a buffer to which the name of 
 *                     the first macro found will be written.
 *     pStartRef    -- (int *) Points to an integer buffer to which the
 *                     relative position of the first character of the 
 *                     macro reference within the string will be written.
 *     pEndRef      -- (int *) Points to an integer buffer to which the
 *                     relative position of the last character of the 
 *                     macro reference within the string will be written.
 *     ppArg         -- (char **) Address of a pointer known to the calling
 *                     function. ExtractMacro will use it to locate a 
 *                     malloced buffer containing the argument string (if
 *                     any) associated with the macro. If there is no 
 *                     argument, *ppArg will be set to NULL.
 *
 * RETURN VALUE: integer. Also see PARAMETERS above.
 *     0 --> No macro reference was found
 *     1 --> A reference was found
 *    -1 --> Some error prevented a macro reference from being processed. 
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: A string containing no macro references is given. POST: An 
 *        'empty' sting is put into the macro name buffer and the indexes 
 *        for the start and end of the (nonexistent) macro reference are 
 *        each set to -1. The function's return value is 0.
 *     2. PRE: A string containing one or more valid macro references is 
 *        given. POST: the first macro reference is located; the relative
 *        positions of the first ('$') and last (')') characters of the
 *        reference are returned through the calling arguments; and the
 *        name of the found macro is stored in the buffer to which pMacro-
 *        Name points. The function's return value is 1.
 *     3. PRE: A string is given in which the first macro reference is bad,
 *        i.e., the macro name is longer than the allowable limit, or the
 *        reference has a missing closing parenthsis (')'). POST: Error
 *        messages are output via macUserError(), and an indicator 
 *        (CompilerFailure) that there's been a serious compiler error is
 *        set. The function's return value is -1.
 *
 * NOTES:
 *     1. Ignores anything between the ']' and the ')' in a macro reference
 *        with argument.
 *
 * REVISION HISTORY:
 *   09/14/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
int macExtractMacro( char *pString, char *pMacroName, unsigned int *pStartRef, 
				 unsigned int *pEndRef, char **ppArg )
{
	unsigned int length, i=0, j=0;
	unsigned int nameLength = 0; /* the length of the name of the macro */
	unsigned int 
		startArg = 0,  /* the index of the first character of the argument
					    * string */
		argLength = 0; /* length of the optional argument string */

	length = strlen( pString );
	*ppArg = (char *) NULL; /* In case none is found prior to return */

	/* Scan string for start of a macro reference: '$('  */
	while
		( 
		   (pString[i] != '$' || pString[i+1] != '(') && 
		   i < length 
		)i++;

	if( i < length ) {
		/* We've found the start of a reference */
		*pStartRef = i;
		i = i + 2; /* index for first char. of name */

		/* Find the end of this macro name...*/
		while ( pString[i] != ')' && pString[i] != '[' && i < length ) i++;
		if( i < length ) {
			/* We've found the end of the reference or the start of an argument*/
			if( pString[i] == ')'  ) { 
				/* It's the end of the reference...*/
				*pEndRef = i;
				/* Compute the length of the macro name...*/
				nameLength = *pEndRef - *pStartRef - 2;
			} else { /* there's an argument */
				unsigned int endName = i - 1; /* index of last character of the
											   * macro name */
				startArg = i + 1; /* index of the first character of the index */
				/* Compute the length of the macro name: */
				nameLength = endName - (*pStartRef + 1);
				/* Find the argument terminator: */
				while( pString[i] != ']' && i < length ) i++;
				if( i < length ) {
					/* We've found the end of the argument */
					unsigned int endArg = i; /* index of the agrument terminator */
					/* Compute the number of characters in the arg string: */
					argLength = (endArg - 1) - (startArg - 1); 
				} else {
					/* An argument spec was started but never finished...*/
					char msg[256];
					sprintf(msg, "Macro argument started but never finished."
						" Bad string follows:\n\t%s", pString );
					macUserError( "ExtractMacro", "Unknown", msg );
					return -1;
				}

				/* Finally, find the end of the entire reference: */
				while( pString[i] != ')' && i < length ) i++;
				if( i < length ) {
					/* We've found the end of the whole reference...*/
					*pEndRef = i;
				} else {
					/* A reference was started but never finished */
					char msg[256];
					sprintf(msg, "Macro reference started but never finished"
					". Bad string follows:\n\t%s", pString );
					macUserError( "ExtractMacro", "Unknown", msg );
					return -1;
				}
			}
		} else {
			/* A reference was started but never finished */
			char msg[256];
			sprintf(msg, "Macro reference started but never finished"
				". Bad string follows:\n\t%s", pString );
			macUserError( "ExtractMacro", "Unknown", msg );
			return -1;
		}
	} else {
		/* It contains no macro references. */
		*pStartRef = -1;
		*pEndRef   = -1;
		*pMacroName = '\0';
		return 0;
	}

	/* Check for long names... */
	if( nameLength > MAX_MACRO_NAME ) {
		/* The macro's name is too long */
		macUserError( "ExtractMacro", "Unknown", 
			"Name exceeds the length limit." );
		*pStartRef = -1;
		*pEndRef   = -1;
		*pMacroName = '\0';
		return -1;
	}

	/* Copy the macro name into the buffer...*/
	for( i = *pStartRef + 2, j=0; j < nameLength; i++, j++ ) 
		pMacroName[j] = pString[i];
	pMacroName[j] = '\0'; /* The string terminator. */

	/* Extract the argument string, if any: */
	if( argLength > 0 ) {
		char *pBuf = NULL;
		pBuf = (char *) malloc( argLength * sizeof(char) + 1 );
		*ppArg = pBuf;
		for( i = startArg, j=0; j < argLength; i++, j++ )
			pBuf[j] = pString[i];
		pBuf[j] = '\0';
	} else *ppArg = NULL;

	/* Normal return...*/
	return 1;
} /* end macExtractMacro */

/*****************************************************************************
 * 
 * NAME: macFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void macFatalError( char *functionName, char *msg )
{
	if( ufmCONTROLS.debug )
		fprintf( ufmCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	macUserWarning( "macFatalError", 
		     "Fatal programming error is causing the program to shut down." );
	exit(-1);

} /* end macFatalError */
/*****************************************************************************
 * 
 * NAME: macUserWarning
 *
 * DESCRIPTION: Reports warnings if the VERBOSE flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void macUserWarning( char *functionName, char *msg )
{
	if( !ufmCONTROLS.verbose ) return;

    fprintf( ufmCONTROLS.pCompilerErrors, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end macUserWarning */

/* --------------------------------------------------------------------------*/

/*****************************************************************************
 * 
 * NAME: macUserError
 *
 * DESCRIPTION: Reports fatal errors and sets CompilerFailure.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     pMacro       -- (char *) pointer to the name (if known) of the macro
 *                     being expanded.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pCompilerErrors -- FILE The output file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void macUserError( char *functionName, char *pMacro, char *msg )
{
    fprintf( ufmCONTROLS.pCompilerErrors, "%s: Bad Macro = %s - %s\n", functionName, pMacro, msg );
	ufmCONTROLS.CompilerFailure = 1;

} /* end macUserError */

/* --------------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: macExpandString
 *
 * DESCRIPTION: Creates a new string in which the characters between
 *		startRef and endRef in the old string have been replaced by the string
 *      to which pMacroString points.
 *
 * PARAMETERS:  
 *     pOldString -- (char *) Points at string containing a macro reference
 *                   that's to be replaced.
 *     pMacroString -- (char *) Points at the string that is to go in place
 *                      of the macro reference.
 *     startRef -- int The index in pOldString at which the macro reference
 *                 starts (i.e., the relative position of the '$' character).
 *     endRef -- int The index in pOldString at which the macro reference
 *                 ends (i.e., the relative position of the ')' character).
 *
 * RETURN VALUE: pointer to a new string containing the expanded version
 *		of pOldString.
 *     
 * NOTES:
 *     <<description>>
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
static char *macExpandString( char *pOldString, char *pMacroString,
							 unsigned int startRef, unsigned int endRef )
{
	char *pNewString;
	unsigned int sizeString;
	unsigned i, j, k;

	/* The size of the new string buffer will be the size of the old one less
	 * the size of the macro reference plus the length of the replacement 
	 * string plus 1 (for the terminator). 
	 */
	if( !pOldString ) macFatalError( "macExpandString", 
		"NULL pointer. to old string." );
	if( !pMacroString ) macFatalError( "macExpandString", 
		"NULL pointer. to substitution string." );
	assert( endRef > startRef );

	sizeString = strlen(pOldString) - ( endRef - (startRef-1) )
		                                          + strlen(pMacroString) + 1;

	pNewString = malloc( sizeString );
	assert( pNewString);

	/* Copy first part of old string */
	for( i=0, j=0; i <  startRef; i++, j++ ) pNewString[j] = pOldString[i];

	/* Copy the macro substitution string */
	for( k=0; k < strlen(pMacroString); k++, j++) 
		                                     pNewString[j] = pMacroString[k];
    /* Copy the last part of the old string */
	for( i=endRef+1; i < strlen(pOldString); i++, j++ )
											   pNewString[j] = pOldString[i];

	pNewString[j] = '\0'; /* add the terminator */

	return pNewString;

} /* end macExpandString */

/*****************************************************************************
 * 
 * NAME: macChildList
 *
 * DESCRIPTION: Creates a string whose value is a comma delimited list of the
 * unique names of all the children of the ODL component to which *pCompo
 * refers. In other words, it builds a ListSpec string for the children
 * (including parameters) of the given component.
 *
 *
 * PARAMETERS:  
 *     pCompo -- (struct ODL_Node*) The component for which the list spec
 *			     of children is desired.
 *
 * RETURN VALUE:
 *     macChildList -- (char *) Comma delimited list of unique names of the
 *               child components of *pCompo.
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * REVISION HISTORY:
 *   09/30/98  Ken Jarva -- Original
 *   10/06/98  Ken Jarva -- rewrote to handle components having multiple
 *                          children of the same name.
 *
 ****************************************************************************/
char *macChildList( struct ODL_Node *pCompo )
{
	AGGREGATE pAgg, pNextAgg;
	PARAMETER pParm;
	lstLIST pChildren;
	unsigned int totlen = 0;
	char *pName = (char *) NULL;
	char *pListSpec = (char *) NULL;

	pChildren = lstInitList();
	assert( pCompo );
	if( pCompo->node_kind != AGGREGATE_NODE ) return strdup("");
	else {

		/* Build a list of all the unique names of the child aggregates: */
		pAgg = pCompo->node_type.aggregate;
		assert( pAgg );
		pNextAgg = pAgg->first_child;
		while( pNextAgg ){
			assert( pNextAgg->name );
			pName = pNextAgg->name;
			if( !lstIsStrInList(pChildren, pName) ){
				lstAddLast( pChildren, pName );
				totlen += ( strlen(pName) + 2 );
			}
			pNextAgg = pNextAgg->right_sibling;
		}

		/* Now do the same for the child parameters: */
		pParm = pAgg->first_parameter;
		while( pParm ){
			assert( pParm->name );
			pName = pParm->name;
			if( !lstIsStrInList(pChildren, pName) ){
				lstAddLast( pChildren, pName );
				totlen += ( strlen(pName) + 2 );
			}
			pParm = pParm->right_sibling;
		}

		/* Protect against childless components...*/
		if( totlen == 0 ) return strdup( "" );

		/* Build the return string: */
		pListSpec = malloc( totlen * sizeof(char) + 1);
		*pListSpec = '\0'; /* terminate it for strcat */
		pName = lstIterator( pChildren );
		while( pName ) {
			strcat( pListSpec, pName );
			strcat( pListSpec, ", " );
			pName = lstIterator( pChildren );
		}
		lstFreeList( pChildren );

		/* Clean up the end */
		pListSpec[ strlen(pListSpec) - 2 ] = '\0';
	}
	return pListSpec;
} /* end macChildList */

/*****************************************************************************
 * 
 * NAME: macDebugMsg
 *
 * DESCRIPTION: Reports debug messages.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void macDebugMsg( char *functionName, char *msg )
{
	if( ufmCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( ufmCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(ufmCONTROLS.pDebugFile);
	}
	return;

} /* end macDebugMsg */

/*****************************************************************************
 * 
 * NAME: macGetMacroString
 *
 * DESCRIPTION: Gets the translation string for the given macro. This 
 *              function handles both user-defined & predefined macros. 
 *
 * PARAMETERS:  
 *     pMacro -- (char *) The macro for which the translation string is
 *                        desired.
 *	   pArg   -- (char *) The argument string, if any associated with
 *                        the given macro.
 *
 * RETURN VALUE:
 *     macGetMacroString -- (char *) The translation string for the given
 *                                   macro.
 *     
 * GLOBAL VARIABLES: 
 *     pCCStack -- lstLIST The Current Component Stack.
 *	
 * NOTES:
 *		1.  Only the _VALUE macro currently uses an argument. Arguments
 *          given with other macros are ignored.
 *
 * REVISION HISTORY:
 *   09/29/98  Ken Jarva  -- Original
 *   10/06/98  Ken Jarva  -- Cleanup & added code for _CURRCOMP.
 *   01/20/99  Ken Jarva  -- Added processing for _VALUE.
 *   02/02/99  Ken Jarva  -- Completed code for handling arguments of _VALUE.
 *
 ****************************************************************************/
char *macGetMacroString( char *pMacro, char *pArg )
{
	char *pSubStr;
	struct ODL_Node *pCC;
	char *pTransString, *pName;

    /* Try to look up the macro in the table of user-defined macros: */
	pSubStr = (char *)sltLookUp( pMacroLUT, pMacro );

	/* Get the current component */
	if( lstNumEl(pCCStack) > 0) {
		/* The current component stack should have at least one
		 * element. */
		pCC = (struct ODL_Node *) pCCStack->first->item;
		assert( pCC );
	} else {
		/* Anytime macros are being resolved, there should be a 'current
		 * component. */
		char msg[256];
		sprintf( msg, "There is no current component for macro: %s", pMacro );
		macDebugMsg( "macGetMacroString", msg );
		pCC = NULL;
	}

	if( pSubStr != (char *) NULL ) {
		/* The macro is in the lookup table */
		return strdup( pSubStr );
	} else if( strcmp(pMacro, "_CHILDREN") == 0 ) {
		/* Predefined macro: all the children of the current
		 * component...*/
		assert( pCC );
		return macChildList( pCC );
	} else if( strcmp(pMacro, "_CURRCOMP") == 0 ) {
		/* Predefined macro: the name of the current component,
		 *     untranslated. */
		assert( pCC );
		pName = odlComponentName( pCC );
		assert( pName );
		return strdup( pName );
	} else if( strcmp(pMacro, "_IDTRANS") == 0 ){
		/* Predefined macro: the translated version of the
		 * current component name. */
		assert( pCC );
		pName = odlComponentName( pCC );
		assert( pName );
		pTransString = (char *)sltLookUp( pIDTranslationLUT, pName );
		/* Duplicate the name */
		if( pTransString ) {
			return strdup( pTransString );
		} else{
			return strdup( pName );
		}
	} else if( strcmp(pMacro, "_VALUE") == 0 ){
		/* process the _VALUE predefined macro with or without an argument */
		if( pArg ) {
			/* The predefined macro has an argument...*/
			lstLIST
				pLists,   /* list of lists containing parameter values */
				pValList, /* a list of values (as strings) of all the parameters
						   * given in pRefList */
				pParmVals, /* list of the values for a particular parameter */
				pRefList; /* a list of references constructed by ListExpander */
			char *pString = strdup("");

			/* Convert the argument into a list of one or more ptrs to ODL parms: */
			pRefList = ListExpander( *pCC, pArg );

			/* Get a list of lists, one list for each of the referenced parms:*/
			pLists = odlListParmVals( pRefList );
			pRefList = lstFreeList( pRefList );

			/* Initialize the single list that will contain all the values from
			 * all the parameters: */
			pValList = lstInitList();

			/* Build the one long list of the values in the lists of pLists: */
			lstStopIteration( pLists );
			  pParmVals = lstIterator( pLists ); /* for the 1st parm referenced */
			  while( pParmVals ) {
				lstAppendTo( pValList, pParmVals );
				pParmVals = lstFreeList( pParmVals );
				pParmVals = lstIterator( pLists ); /* for the next parm referenced */
			  }
			lstStopIteration( pLists );

			/* Create a comma-delimited string with all the parameter values: */
			pString = lstJoinList( ", ", pValList );

			/* Clean up and return: */
			pValList = lstFreeList( pValList );
			pLists = lstFreeList( pLists );
			return strdup( pString );

		} else if( pCC->node_kind != PARAMETER_NODE ) {
			/* It's a simple value reference, but the current component isn't an
			 * ODL parameter...*/
			char msg[256];
			sprintf( msg, "_VALUE reference in other than a parameter component." );
			macUserWarning( "macGetMacroString", msg );
			return strdup("");
		} else /* wants the value for the current component which is a paramter */
			return strdup("./");
	} else {
		/* The macro hasn't been defined...*/
		char msg[256];
		sprintf( msg, "Undefined macro: %s\n", pMacro );
		macUserWarning( "macGetMacroString", msg );
		return strdup("");
	}
} /* end macGetMacroString */

