/* FILE: sltLUT.c
 *
 * PROGRAM: Utility module
 *
 * DESCRIPTION: This file contains functions that manipulate lookup
 *              tables.
 *
 * RCS INFO:
 *
 *      $Id: sltLUT.c,v 1.1 1999/01/06 21:08:32 ToolsDev Exp $
 *
 *      $Log: sltLUT.c,v $
 *      Revision 1.1  1999/01/06 21:08:32  ToolsDev
 *      Initial revision from prototype baseline.
 *
 *
 * NOTICE: this file was constructed from a template that supplied a
 *         control structure and some simple functions for reporting
 *         errors and debug information. See the static struct 
 *         sltCONTROLS below.
*/

/* INCLUDED FILES:
*/
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "sltLUT.h"
#include "strUtilities.h"
/* ---------------------------------------------------------------------*/
/* MACRO DEFINITIONS */

#define VERBOSE_DEBUG 0  /* Set to 1 for maximal debug output but also
                          * see sltCONTROLS.debug.

/* ---------------------------------------------------------------------*/
/* GLOBAL VARIABLES: the following variables may be used in other files.
 */
	/* NONE */

/* ---------------------------------------------------------------------*/
/* EXTERNAL VARIABLES: the following variables are defined in other files.
*/
	/* NONE */

/* ---------------------------------------------------------------------*/
/* STATIC VARIABLES: the following variables are for the exclusive use
 *     of the functions defined within this file.
*/
struct sltControls {
    /* Controls the message writing functions given in the source file
	 * template (e.g., sltDebugMsg().  This structure is
	 * imediately followed by its default settings. Typically the
	 * initailization function sltInitialization is invoked by
	 * the calling application to override these settings. */
	int debug;
	int verbose;
	int noExit;
	FILE *pDebugFile;
	FILE *pUserMessages;
} sltCONTROLS =
{
	1,
	1,
	0,
	stderr,
	stderr
};

/* The following struct is the entry in a lstList representing a LUT. */
struct sltEntry{
	char *pIndex;
	void *pItem;
	char IsA[6];  /* set to "ENTRY" by sltAdd(). */
};
typedef struct sltEntry *sltENTRY;


/* ---------------------------------------------------------------------*/

/* STATIC PROTOTYPES: the following prototypes are for functions defined
 *     and used exclusively within this file.
*/
static void sltFatalError( char *functionName, char *msg );
static void sltUserWarning( char *functionName, char *msg );
static void sltDebugMsg( char *functionName, char *msg );
static sltENTRY sltEntryItem( void *pEntry, char **pIndex );

/* ---------------------------------------------------------------------*/
/*****************************************************************************
 * 
 * NAME: sltInitialization
 *
 * DESCRIPTION: Initializes the control structurs for this module.
 *
 * PARAMETERS:  
 *     debugFlag     -- (int) Control to allow debug print to be output. Set
 *                      to 1 for ouput; 0 otherwise.
 *     verboseFlag   -- (int) Control to allow verbose messages for the user
 *                      to be output. Set to 1 for ouput; 0 otherwise.
 *	   noExit        -- (int) Control to prevent the functions within this
 *                      module from exiting the program.  Set to 1 and all
 *                      functions return to the caller, including 
 *                      sltFatalError(...).
 *     pDebugFile    -- (FILE *) Pointer to the debug file to which debug
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 *     pUserMessages -- (FILE *) Pointer to the debug file to which user
 *                      messages (if allowed) will be written.  If passed
 *                      as NULL, stderr will be used if necessary.  Only
 *                      NULL or an active FILE pointer should be passed.
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct sltCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *     verbose    -- (int) control allowing verbose messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void sltInitialization( int debugFlag, int verboseFlag, int noExit,
	FILE *pDebugFile, FILE *pUserMessages )
{
	sltCONTROLS.debug = debugFlag;
	sltCONTROLS.verbose = verboseFlag;
	sltCONTROLS.noExit = noExit;
	if( pDebugFile ) sltCONTROLS.pDebugFile = pDebugFile;
	if( pUserMessages ) sltCONTROLS.pUserMessages = pUserMessages;
	return;
} /* end sltInitialization */

/*****************************************************************************
 * 
 * NAME: sltFatalError
 *
 * DESCRIPTION: Reports fatal errors and shuts down the program. Messages
 *              are written only if the debug control is set. If noExit is
 *              set the funtion reutrns to the caller; if not the program
 *              exited with a -1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct sltCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *     noExit     -- (int) control preventing the function from exiting the
 *                   program.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void sltFatalError( char *functionName, char *msg )
{
	if( sltCONTROLS.debug )
		fprintf( sltCONTROLS.pDebugFile, 
		                              "%s: FATAL - %s\n", functionName, msg );
	if( !sltCONTROLS.noExit ) {
		sltUserWarning( "sltFatalError", 
		     "Fatal programming error is causing the program to shut down." );
		exit(-1);
	}
	else return;

} /* end sltFatalError */

/*****************************************************************************
 * 
 * NAME: sltUserWarning
 *
 * DESCRIPTION: Reports warnings to the user if the verbose flag is set.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct sltCONTROLS:
 *     pUserMessages -- (FILE) The debug file to which error messages are to 
 *                             be written.
 *     verbose       -- (int)  Control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void sltUserWarning( char *functionName, char *msg )
{
	if( !sltCONTROLS.verbose ) return;

    fprintf( sltCONTROLS.pUserMessages, 
		                             "%s: WARNING - %s\n", functionName, msg );
	return;

} /* end sltUserWarning */
/*****************************************************************************
 * 
 * NAME: sltDebugMsg
 *
 * DESCRIPTION: Reports debug messages for the maintenance programmer if the
 *              debug flag is set and VERBOSE_DEBUG macro is 1.
 *
 * PARAMETERS:  
 *     functionName -- (char *) pointer to a string containing the name of
 *                      the function that trapped the fatal error.
 *     msg          -- (char *) pointer to a buffer containing a description
 *                      of the error that occurred.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES:
 *   struct sltCONTROLS:
 *     pDebugFile -- FILE The debug file to which error messages are to be 
 *                   written.
 *     debug      -- (int) control allowing debug messages to be output.
 *	
 * REVISION HISTORY:
 *   09/15/98  Ken Jarva  -- original
 *   11/17/98  Ken Jarva  -- included as a standard function in the source
 *                           file template.
 *
 ****************************************************************************/

void sltDebugMsg( char *functionName, char *msg )
{
	if( sltCONTROLS.debug && VERBOSE_DEBUG ){
		fprintf( sltCONTROLS.pDebugFile, 
		                              "%s: INFO - %s\n", functionName, msg );
	    fflush(sltCONTROLS.pDebugFile);
	}
	return;

} /* end sltDebugMsg */
 
/*****************************************************************************
 * 
 * NAME: sltInitLUT
 *
 * DESCRIPTION: Initializes a lookup table.
 *
 * PARAMETERS:  None
 *
 * RETURN VALUE: pointer to lookup table
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. A LUT is just a special purpose list. This may change.
 *
 * REVISION HISTORY:
 *   09/12/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/
sltLUT sltInitLUT( void )
{
	lstLIST pList;
	
	pList = lstInitList();
	return (sltLUT) pList;

} /* end sltInitLUT */
/*****************************************************************************
 * 
 * NAME: sltFreeLUT
 *
 * DESCRIPTION: Frees the lookup table: that is, it frees the underlying
 *		list root and each of the list nodes. It also frees the string
 *      used to index the LUT.  The arbitrary data item to which an index
 *      refers is not freed.
 *
 * PARAMETERS:  
 *     pLUT -- sltLUT Pointer to the LUT that's to be freed.
 *
 * RETURN VALUE: (sltLUT) NULL
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Iterates the underlying list so it can free the index strings.
 *     2. Uses lstFreeList to free the underlying list.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

sltLUT sltFreeLUT(sltLUT pLUT)
{
	lstLIST pList;
	sltENTRY pEntry;

	pList = (lstLIST) pLUT;

	lstStopIteration( pList ); /* precaution */

	/* Visit each entry in the table; free its index string and
	 * then the entry struct itself.  The associated data item is
	 * NOT freed.
	 */
	pEntry = (sltENTRY) lstIterator( pList );
	while( pEntry != (sltENTRY) NULL ) {
		if( pEntry->pIndex != (char *)NULL ) free( pEntry->pIndex );
		free( pEntry );
		pEntry = (sltENTRY) lstIterator( pList ); 
	}
	return lstFreeList( (lstLIST) pLUT );
} /* end sltFreeLUT */

/*****************************************************************************
 * 
 * NAME: sltAdd
 *
 * DESCRIPTION: Makes an entry into the given LUT.
 *
 * PARAMETERS:  
 *     indexString -- char * The index by which the data item shall be known.
 *     pItem       -- void * Pointer to the data item.
 *
 * RETURN VALUE: None
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     ***NOTICE*** the index string and the assoicated item are treated
 *     differently: the index string is duplicated and the copy is then used
 *     by the LUT; so sltAdd could be called with something like
 *                 sltAdd( pLUT, "index1", pWhatever );
 *     without having to worry about the index evaporating as control passes
 *     uppwards.  The data item, which is of unknown complexity, must exist
 *     at least as long as does the LUT.  An independent copy is NOT saved.
 *     All we have is a pointer to the thing, and as long as the LUT exists,
 *     the thing pointed to must exist.  It doen't necessarily have to exist
 *     on the heap, but it is somewhat dangerous to have pointers on the heap
 *     pointing at things that're on the stack.
 *
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva             Raytheon
 *
 ****************************************************************************/

void sltAdd(sltLUT pLUT, char* indexString, void *pItem )
{
	sltENTRY pEntry;

	/* Make a copy of the index string on the heap */
	pEntry = malloc( sizeof(struct sltEntry) );
	assert( pEntry );

	/* Set up the entry structure... */
	strcpy( pEntry->IsA, "ENTRY" );
	pEntry->pIndex = strdup( indexString );
	pEntry->pItem = pItem;

	/* Add the entry structure to the underlying list...*/
	lstAddLast( (lstLIST) pLUT, pEntry );

	return;
} /* end sltAdd */

/*****************************************************************************
 * 
 * NAME: sltLookUp
 *
 * DESCRIPTION: Searches the LUT for an entry matching the given indexString
 *		and returns a pointer to the assoicated data item if a match is made
 *      and a NULL pointer otherwise.
 *
 * PARAMETERS:  
 *     pLUT        -- sltLUT The lookup table to be searched.
 *     indexString -- (char *) Points at the string to be matched.
 *
 * RETURN VALUE: (void *) The data item assoicated with the given index string.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Generally it is not considered an error if the index string cannont be
 *        matched with an entry in the LUT; however, a warning will be issued on
 *        such occasions, provided the verbosity controls allow.
 *     2. Note the use of the list interator. Be sure that lstStopIteration is
 *        called prior to exiting this function.
 *
 * REVISION HISTORY:
 *   09/11/98  Ken Jarva -- Original 
 *   10/29/98  Ken Jarva -- buried warning; added note 2.
 *
 ****************************************************************************/
void *sltLookUp( sltLUT pLUT, char *indexString )
{
	lstLIST pList;
	sltENTRY pEntry;
	char msg[256];

	pList = (lstLIST) pLUT; /* a LUT is just a list */

	lstStopIteration( pList ); /* mere precaution */

	/* Search the LUT for an entry with an index that matchecs
	 * indexString. */
	pEntry = (sltENTRY) lstIterator( pList );
	while( pEntry != (sltENTRY) NULL ) {
		if( strToUpStrcmp(pEntry->pIndex, indexString) == 0 ){
			/* We've found a match. */
			lstStopIteration( pList );
			return pEntry->pItem;
		}
		/* No match yet; get the next entry */
		pEntry = (sltENTRY) lstIterator( pList );
	}

	/* Failed to match given string index...*/
	lstStopIteration( pList );
	sprintf( msg, "failed to find entry for %s\n", indexString );
	if(VERBOSE_DEBUG) sltUserWarning( "sltLookUp", msg );
	return (void *)NULL;

} /* end sltLookUp */

/*****************************************************************************
 * 
 * NAME: sltMatchList
 *
 * DESCRIPTION: Searches the LUT for entries which include the given 
 *      indexString as the first strlen( indexString ) characeters. It returns
 *      a list of indexes from the LUT for which this is the case.
 *
 * PARAMETERS:  
 *     pLUT        -- sltLUT The lookup table to be searched.
 *     indexString -- (char *) Points at the string to be matched.
 *
 * RETURN VALUE: (void *) The data item assoicated with the given index string.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. Note the use of the list interator. Be sure that lstStopIteration is
 *        called prior to exiting this function.
 *
 * REVISION HISTORY:
 *   11/23/98  Ken Jarva -- Original 
 *
 ****************************************************************************/
lstLIST sltMatchList( sltLUT pLUT, char *indexString )
{
	lstLIST pList, pNames;
	sltENTRY pEntry;

	pList = (lstLIST) pLUT; /* a LUT is just a list */
	pNames = lstInitList(); /* The indexes that start with indexString */

	lstStopIteration( pList ); /* mere precaution */

	/* Search the LUT for an entry with an index that matches
	 * indexString. */
	pEntry = (sltENTRY) lstIterator( pList );
	while( pEntry != (sltENTRY) NULL ) {
		char *pFirstPart = strdup( pEntry->pIndex );
		unsigned int n = strlen(indexString);
		if( strlen(pFirstPart) >= n ){
			pFirstPart[n] = '\0';
			if( strToUpStrcmp(pFirstPart, indexString) == 0 ){
				/* We've found a match. */
				lstAddLast( pNames, pEntry->pIndex );
			}
		}
		/* No match yet; get the next entry */
		pEntry = (sltENTRY) lstIterator( pList );
		free( pFirstPart );
	}

	/* Failed to match given string index...*/
	lstStopIteration( pList );

	if( lstNumEl( pNames ) > 1 ) {
		char msg[256];
		sprintf( msg, "Have %d defs for %s\n", 
			lstNumEl(pNames), indexString );
		sltDebugMsg("sltMatchList", msg );
	}

	return pNames;

} /* end sltMatchList */

/*****************************************************************************
 * 
 * NAME: sltAddOrReplace
 *
 * DESCRIPTION: If an entry for the given index string already exists within
 *              the LUT, this function replaces it with *pItem; otherwise, it
 *              adds pItem to the LUT.
 *
 * PARAMETERS:  
 *     pLUT         -- (sltLUT) The lookup table that's to be manipulated.
 *	   pIndexString -- (char *) Index into the lookup table.
 *     pItem        -- (void *) The item that is to go into the LUT.
 *
 * RETURN VALUE:
 *     sltAddOrReplace -- (void *) Points at the original item that was indexed
 *                                 by *pIndexString if there was one. NULL if
 *                                 there wasn't.
 *     
 * GLOBAL VARIABLES: None
 *	
 * PRE- & POST-CONDITIONS:
 *     1. PRE: a lookup table (LUT) is given in which there exists an entry
 *        indexed by the string to which pIndexString points. POST: the 
 *        original entry in the LUT is replaced with pItem. A (void) pointer
 *        to the thing originally listed is returned as the function's value.
 *     2. PRE: A LUT is given in which no entry for *pIndexString exists.
 *        POST: pItem is added to the LUT indexed by pIndexString.
 *
 * NOTES:
 *     1. The original thing stored in the LUT, if it exists, must be returned
 *        to the caller because we can't know here whether or not it should be
 *        freed.
 *
 * REVISION HISTORY:
 *   10/24/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
void *sltAddOrReplace( sltLUT pLUT, char *pIndexString, void *pItem )
{
	void *pEntry, *pWhatever;
	char *pIndex;
	void *pOriginalItem = (void *) NULL;

	/* Protect against nonesensical input: */
	if( !pLUT ) sltFatalError("sltAddOrReplace", 
		"Was given NULL pointer to LUT." );
	if( !pIndexString )  sltFatalError("sltAddOrReplace", 
		"Was given NULL pointer to index string." );

	/* Step through the entries and try to find an occurrence
	 * of *pIndexString, and if it's there replace it: */
	lstStopIteration( pLUT );
		pEntry = lstIterator( pLUT );
		while( pEntry ) {
			pWhatever = sltEntryItem( pEntry, &pIndex );
			if( strToUpStrcmp(pIndexString, pIndex) == 0 ) {
				pOriginalItem = lstReplaceCurrent( pLUT, pItem);
				pOriginalItem = pWhatever;
				break; /* Can't return 'til we stop the iteration...*/
			}
			pEntry = lstIterator( pLUT );
		}
	lstStopIteration( pLUT );

	/* It there was nothing stored by that name, just add it: */
	if( !pOriginalItem) sltAdd( pLUT, pIndexString, pItem );
	
	return pOriginalItem;

} /* end sltAddOrReplace */

/*****************************************************************************
 * 
 * NAME: sltEntryItem
 *
 * DESCRIPTION: Given a entry struct for a string indexed lookup table, this 
 *     function returns the associated void pointer. It also returns the index 
 *     string through pIndex. 
 *
 *
 * PARAMETERS:  
 *     pEntry  -- (void *) Pointer to the sltEntry struct from which the 
 *                         information is to be extracted.
 *     ppIndex -- (char **) Pointer to a pointer to a buffer into which the
 *                         index string is to be written.
 *
 * RETURN VALUE:
 *     sltEntryItem -- (void *) Pointer to whatever was stored in the LUT.
 *     
 * GLOBAL VARIABLES: None
 *	
 *
 * NOTES:
 *     1. *pEntry is checked to make sure it is in fact an SLT entry.
 *     2. Asserts that neiter pEntry nor pIndex is NULL.
 *
 * REVISION HISTORY:
 *   10/07/98  Ken Jarva  --  Original
 *
 ****************************************************************************/
sltENTRY sltEntryItem( void *pEntry, char **ppIndex )
{
	sltENTRY pEnt;

	assert( pEntry );
	assert( ppIndex );
	pEnt = (sltENTRY) pEntry;

	if( strToUpStrcmp(pEnt->IsA, "ENTRY") != 0 ) return NULL;
	else {
		*ppIndex = pEnt->pIndex;
	}
	return pEnt->pItem;

} /* end sltEntryItem */

/*****************************************************************************
 * 
 * NAME: sltKeys
 *
 * DESCRIPTION: Returns the indexes (aka keys) of the LUT. Modeled
 *              on the Perl function for hashes.
 *
 * PARAMETERS: 
 *     pLUT -- (sltLUT) Pointer to a string-indexed look up table.
 *
 * RETURN VALUE:
 *     sltKeys -- (lstLIST) List of the indexes (similar to hash keys in
 *                          Perl) used in the given lookup table.
 *     
 * GLOBAL VARIABLES: None
 *	
 * NOTES:
 *     1. This function must always return a pointer to a valid lstList
 *        struct.
 *     2. The strings to which the lstList entries will point are the
 *        indexes in the LUT. ***They should NOT be freed when (if) *pList
 *        is freed.
 *
 * REVISION HISTORY:
 *   12/15/98  Ken Jarva  --  Original
 *
 ****************************************************************************/

lstLIST sltKeys(sltLUT pLUT)
{
	lstLIST pList; /* list to be returned */
	void 
		*pEntry,	   /* an entry from *pLUT */
		*pItem;     /* the item stored in *pLUT; we're not really interested
				    * in this; we just want its index */
	char *pIndex;  /* the index string associated with a particular entry
				    * of the LUT */

	pList = lstInitList(); /* Initialize list of indexes (i.e., keys ) */
	if( !pLUT ) return pList; /* precaution */

	lstStopIteration( (lstLIST) pLUT ); /* treat the LUT like a lstList */
	  pEntry = lstIterator( (lstLIST) pLUT );
	  while( pEntry ) {
		pItem = sltEntryItem( pEntry, &pIndex ); /* pIndex is what we're
												  * after. */
		lstAddLast( pList, pIndex );
	    pEntry = lstIterator( (lstLIST) pLUT );
	  }
	lstStopIteration( (lstLIST) pLUT ); 
	return pList;

} /* end sltKeys */




